from typing import Optional, Set, Tuple, List
import spimdisasm
from ..segment import Segment
from ...util import log, options, symbols

from .data import CommonSegData

from ...disassembler.disassembler_section import (
    DisassemblerSection,
    make_rodata_section,
)


class CommonSegRodata(CommonSegData):
    def get_linker_section(self) -> str:
        return ".rodata"

    def get_section_flags(self) -> Optional[str]:
        return "a"

    @staticmethod
    def is_data() -> bool:
        return False

    @staticmethod
    def is_rodata() -> bool:
        return True

    def get_possible_text_subsegment_for_symbol(
        self, rodata_sym: spimdisasm.mips.symbols.SymbolBase
    ) -> Optional[Tuple[Segment, spimdisasm.common.ContextSymbol]]:
        # Check if this rodata segment does not have a corresponding code file, try to look for one

        if self.sibling is not None or not options.opts.pair_rodata_to_text:
            return None

        if not rodata_sym.shouldMigrate():
            return None

        if len(rodata_sym.contextSym.referenceFunctions) != 1:
            return None

        func = list(rodata_sym.contextSym.referenceFunctions)[0]
        text_segment = self.parent.get_subsegment_for_ram(func.vram)

        if text_segment is None or not text_segment.is_text():
            return None
        return text_segment, func

    def configure_disassembler_section(
        self, disassembler_section: DisassemblerSection
    ) -> None:
        "Allows to configure the section before running the analysis on it"

        section = disassembler_section.get_section()

        # Set rodata string encoding
        # First check the global configuration
        if options.opts.string_encoding is not None:
            section.stringEncoding = options.opts.string_encoding

        # Then check the per-segment configuration in case we want to override the global one
        if self.str_encoding is not None:
            section.stringEncoding = self.str_encoding

    def disassemble_data(self, rom_bytes):
        if self.is_auto_segment:
            return

        if not isinstance(self.rom_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a rom_start. Got '{self.rom_start}'"
            )

        # Supposedly logic error, not user error
        assert isinstance(self.rom_end, int), self.rom_end

        # Supposedly logic error, not user error
        segment_rom_start = self.get_most_parent().rom_start
        assert isinstance(segment_rom_start, int), segment_rom_start

        if not isinstance(self.vram_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a vram address. Got '{self.vram_start}'"
            )

        self.spim_section = make_rodata_section(
            self.rom_start,
            self.rom_end,
            self.vram_start,
            self.name,
            rom_bytes,
            segment_rom_start,
            self.get_exclusive_ram_id(),
        )

        assert self.spim_section is not None

        self.configure_disassembler_section(self.spim_section)

        self.spim_section.analyze()
        self.spim_section.set_comment_offset(self.rom_start)

        possible_text_segments: Set[Segment] = set()

        last_jumptable_addr_remainder = 0
        misaligned_jumptable_offsets: List[int] = []

        for symbol in self.spim_section.get_section().symbolList:
            generated_symbol = symbols.create_symbol_from_spim_symbol(
                self.get_most_parent(), symbol.contextSym, force_in_segment=True
            )
            generated_symbol.linker_section = self.get_linker_section_linksection()

            # Gather symbols found by spimdisasm and create those symbols in splat's side
            for referenced_vram in symbol.referencedVrams:
                context_sym = self.spim_section.get_section().getSymbol(
                    referenced_vram, tryPlusOffset=False
                )
                if context_sym is not None:
                    symbols.create_symbol_from_spim_symbol(
                        self.get_most_parent(), context_sym, force_in_segment=False
                    )

            possible_text = self.get_possible_text_subsegment_for_symbol(symbol)
            if possible_text is not None:
                text_segment, refenceeFunction = possible_text
                if text_segment not in possible_text_segments:
                    print(
                        f"\nRodata segment '{self.name}' may belong to the text segment '{text_segment.name}'"
                    )
                    print(
                        f"    Based on the usage from the function {refenceeFunction.getNameUnquoted()} to the symbol {symbol.getNameUnquoted()}"
                    )
                    possible_text_segments.add(text_segment)

            if options.opts.platform in ("psx", "ps2"):
                if generated_symbol.type == "jtbl":
                    # GCC aligns jumptables to 8, but it doesn't impose alignment restrictions for sections themselves on PSX/PS2.
                    # This means a jumptable may be aligned file-wise, but it may not end up 8-aligned binary-wise.
                    # We can use this as a way to find file splits on PSX/PS2
                    vram_diff = generated_symbol.vram_start - self.vram_start
                    if vram_diff % 8 != last_jumptable_addr_remainder:
                        # Each time the this remainder changes means there's a new file split
                        last_jumptable_addr_remainder = vram_diff % 8

                        misaligned_jumptable_offsets.append(self.rom_start + vram_diff)

        if len(misaligned_jumptable_offsets) > 0:
            print(
                f"\nThe rodata segment '{self.name}' has jumptables that are not aligned properly file-wise, indicating one or more likely file split."
            )
            print(
                "File split suggestions for this segment will follow in config yaml format:"
            )
            for offset in misaligned_jumptable_offsets:
                print(f"      - [0x{offset:X}, {self.type}]")
            print()



#! /usr/bin/env python3

import argparse
import hashlib
from pathlib import Path
import subprocess
import sys
from typing import Optional

from ..util.n64 import find_code_length, rominfo
from ..util.psx import psxexeinfo
from ..util.ps2 import ps2elfinfo
from ..util import log, file_presets, conf


def main(file_path: Path, objcopy: Optional[str]):
    if not file_path.exists():
        sys.exit(f"File {file_path} does not exist ({file_path.absolute()})")
    if file_path.is_dir():
        sys.exit(f"Path {file_path} is a directory ({file_path.absolute()})")

    # Check for N64 ROM
    if file_path.suffix.lower() == ".n64" or file_path.suffix.lower() == ".z64":
        create_n64_config(file_path)
        return

    file_bytes = file_path.read_bytes()

    # Check for PSX executable
    if file_bytes[0:8] == b"PS-X EXE":
        create_psx_config(file_path, file_bytes)
        return

    # Check for ELFs
    if file_bytes[0:4] == b"\x7fELF":
        do_elf(file_path, file_bytes, objcopy)
        return

    log.error(f"create_config does not support the file format of '{file_path}'")


def remove_invalid_path_characters(p: str) -> str:
    invalid_characters = ["<", ">", ":", '"', "/", "\\", "|", "?", "*"]
    for invalid in invalid_characters:
        p = p.replace(invalid, "_")
    return p


def create_n64_config(rom_path: Path):
    rom_bytes = rominfo.read_rom(rom_path)

    rom = rominfo.get_info(rom_path, rom_bytes)
    basename = rom.name.replace(" ", "").lower()
    cleaned_basename = remove_invalid_path_characters(basename)

    header = f"""\
name: {rom.name.title()} ({rom.get_country_name()})
sha1: {rom.sha1}
options:
  basename: {basename}
  target_path: {rom_path.with_suffix(".z64")}
  elf_path: build/{cleaned_basename}.elf
  base_path: .
  platform: n64
  compiler: {rom.compiler}

  # asm_path: asm
  # src_path: src
  # build_path: build
  # create_asm_dependencies: True

  ld_script_path: {cleaned_basename}.ld
  ld_dependencies: True

  find_file_boundaries: True
  header_encoding: {rom.header_encoding}

  o_as_suffix: True
  use_legacy_include_asm: False
  mips_abi_float_regs: o32

  # section_order: [".text", ".data", ".rodata", ".bss"]
  # auto_link_sections: [".data", ".rodata", ".bss"]

  symbol_addrs_path:
    - symbol_addrs.txt
  reloc_addrs_path:
    - reloc_addrs.txt

  # undefined_funcs_auto_path: undefined_funcs_auto.txt
  # undefined_syms_auto_path: undefined_syms_auto.txt

  extensions_path: tools/splat_ext

  # string_encoding: ASCII
  # data_string_encoding: ASCII
  rodata_string_guesser_level: 2
  data_string_guesser_level: 2
  # libultra_symbols: True
  # hardware_regs: True
  # gfx_ucode: # one of [f3d, f3db, f3dex, f3dexb, f3dex2]
"""

    # Start analysing after the entrypoint segment.
    first_section_end = find_code_length.run(
        rom_bytes, 0x1000 + rom.entrypoint_info.segment_size(), rom.entry_point
    )

    extra_message = ""
    if not rom.entrypoint_info.traditional_entrypoint:
        extra_message = " # This game uses a non-traditional entrypoint, meaning splat's analysis may be wrong"

    segments = f"""\
segments:
  - name: header
    type: header
    start: 0x0

  - name: ipl3
    type: bin
    start: 0x40

  - name: entry{extra_message}
    type: code
    start: 0x1000
    vram: 0x{rom.entry_point:X}
    subsegments:
      - [0x1000, hasm]
"""
    if rom.entrypoint_info.data_size is not None:
        segments += f"""\
      - [0x{0x1000 + rom.entrypoint_info.entry_size:X}, data]
"""

    main_rom_start = 0x1000 + rom.entrypoint_info.segment_size()
    segments += f"""\

  - name: main
    type: code
    start: 0x{main_rom_start:X}
    vram: 0x{rom.entry_point + rom.entrypoint_info.segment_size():X}
    follows_vram: entry
"""

    if rom.entrypoint_info.bss_size is not None:
        segments += f"""\
    bss_size: 0x{rom.entrypoint_info.bss_size.value:X}
"""

    segments += f"""\
    subsegments:
      - [0x{main_rom_start:X}, asm]
"""

    if (
        rom.entrypoint_info.bss_size is not None
        and rom.entrypoint_info.bss_start_address is not None
        and first_section_end > main_rom_start
    ):
        bss_start = (
            rom.entrypoint_info.bss_start_address.value - rom.entry_point + 0x1000
        )
        # first_section_end points to the start of data
        segments += f"""\
      - [0x{first_section_end:X}, data]
      - {{ type: bss, vram: 0x{rom.entrypoint_info.bss_start_address.value:08X} }}
"""
        # Point next segment to the detected end of the main one
        first_section_end = bss_start

    if first_section_end > main_rom_start:
        segments += f"""\

  - type: bin
    start: 0x{first_section_end:X}
    follows_vram: main
"""

    segments += f"""\

  - [0x{rom.size:X}]
"""

    out_file = Path(f"{cleaned_basename}.yaml")
    with out_file.open("w", encoding="utf-8", newline="\n") as f:
        print(f"Writing config to {out_file}")
        f.write(header)
        f.write(segments)

    # `file_presets` requires an initialized `opts`.
    # A simple way to do that is to simply load the yaml we just generated.
    conf.load([out_file])
    file_presets.write_all_files()

    # Write reloc_addrs.txt file
    reloc_addrs: list[str] = []

    addresses_info: list[tuple[Optional[rominfo.EntryAddressInfo], str]] = [
        (rom.entrypoint_info.main_address, "main"),
        (rom.entrypoint_info.bss_start_address, "main_BSS_START"),
        (rom.entrypoint_info.bss_size, "main_BSS_SIZE"),
        (rom.entrypoint_info.bss_end_address, "main_BSS_END"),
    ]

    for addr_info, sym_name in addresses_info:
        if addr_info is None:
            continue
        if addr_info.ori:
            # Avoid emitting relocations for `ori`s since `%lo` doesn't support it.
            continue
        if addr_info.rom_hi == addr_info.rom_lo:
            # hi and lo may be the same for the "main" address, i.e. a direct jal.
            continue

        reloc_addrs.append(
            f"rom:0x{addr_info.rom_hi:06X} reloc:MIPS_HI16 symbol:{sym_name}"
        )
        reloc_addrs.append(
            f"rom:0x{addr_info.rom_lo:06X} reloc:MIPS_LO16 symbol:{sym_name}"
        )
        reloc_addrs.append("")

    if (
        rom.entrypoint_info.stack_top is not None
        and not rom.entrypoint_info.stack_top.ori
    ):
        reloc_addrs.append(
            '// This entry corresponds to the "stack top", which is the end of the array used as the stack for the main segment.'
        )
        reloc_addrs.append(
            "// It is commented out because it was not possible to infer what the start of the stack symbol is, so you'll have to figure it out by yourself."
        )
        reloc_addrs.append(
            "// Once you have found it you can properly name it and specify the length of this stack as the addend value here."
        )
        reloc_addrs.append(
            f"// The address of the end of the stack is 0x{rom.entrypoint_info.stack_top.value:08X}."
        )
        reloc_addrs.append(
            f"// A common size for this stack is 0x2000, so try checking for the address 0x{rom.entrypoint_info.stack_top.value - 0x2000:08X}. Note the stack may have a different size."
        )
        reloc_addrs.append(
            f"// rom:0x{rom.entrypoint_info.stack_top.rom_hi:06X} reloc:MIPS_HI16 symbol:main_stack addend:0xXXXX"
        )
        reloc_addrs.append(
            f"// rom:0x{rom.entrypoint_info.stack_top.rom_lo:06X} reloc:MIPS_LO16 symbol:main_stack addend:0xXXXX"
        )
        reloc_addrs.append("")
    if reloc_addrs:
        with Path("reloc_addrs.txt").open("w", encoding="utf-8", newline="\n") as f:
            print("Writing reloc_addrs.txt")
            f.write(
                "// Visit https://github.com/ethteck/splat/wiki/Advanced-Reloc for documentation about this file\n"
            )
            f.write("// entrypoint relocs\n")
            contents = "\n".join(reloc_addrs)
            f.write(contents)

    # Write symbol_addrs.txt file
    symbol_addrs = []
    symbol_addrs.append(f"entrypoint = 0x{rom.entry_point:08X}; // type:func")
    if rom.entrypoint_info.main_address is not None:
        symbol_addrs.append(
            f"main = 0x{rom.entrypoint_info.main_address.value:08X}; // type:func"
        )
    if symbol_addrs:
        symbol_addrs.append("")
        with Path("symbol_addrs.txt").open("w", encoding="utf-8", newline="\n") as f:
            print("Writing symbol_addrs.txt")
            f.write(
                "// Visit https://github.com/ethteck/splat/wiki/Adding-Symbols for documentation about this file\n"
            )
            contents = "\n".join(symbol_addrs)
            f.write(contents)


def create_psx_config(exe_path: Path, exe_bytes: bytes):
    exe = psxexeinfo.PsxExe.get_info(exe_path, exe_bytes)
    basename = exe_path.name.replace(" ", "").lower()
    cleaned_basename = remove_invalid_path_characters(basename)

    header = f"""\
name: {exe_path.name}
sha1: {exe.sha1}
options:
  basename: {basename}
  target_path: {exe_path}
  elf_path: build/{cleaned_basename}.elf
  base_path: .
  platform: psx
  compiler: PSYQ

  # asm_path: asm
  # src_path: src
  # build_path: build
  # create_asm_dependencies: True

  ld_script_path: {cleaned_basename}.ld
  ld_dependencies: True

  find_file_boundaries: False
  gp_value: 0x{exe.initial_gp:08X}
  # ld_gp_expression: main_SCOMMON_START + 0x7FF0

  o_as_suffix: True
  use_legacy_include_asm: False

  section_order: [".rodata", ".text", ".data", ".bss"]
  # auto_link_sections: [".data", ".rodata", ".bss"]

  symbol_addrs_path:
    - symbol_addrs.txt
  reloc_addrs_path:
    - reloc_addrs.txt

  # undefined_funcs_auto_path: undefined_funcs_auto.txt
  # undefined_syms_auto_path: undefined_syms_auto.txt

  extensions_path: tools/splat_ext

  subalign: 2

  string_encoding: ASCII
  data_string_encoding: ASCII
  rodata_string_guesser_level: 2
  data_string_guesser_level: 2

  # Uncomment this line if you need to use the maspsx reorder workaround hack
  # https://github.com/mkst/maspsx?tab=readme-ov-file#include_asm-reordering-workaround-hack
  # include_asm_macro_style: maspsx_hack
"""

    segments = f"""\
segments:
  - name: header
    type: header
    start: 0x0

  - name: main
    type: code
    start: 0x800
    vram: 0x{exe.destination_vram:X}
    # bss_size: Please fill out this value when you figure out the bss size
    subsegments:
"""
    text_offset = exe.text_offset
    if text_offset != 0x800:
        segments += """\
      - [0x800, rodata, 800]
"""
    segments += f"""\
      - [0x{text_offset:X}, asm, {text_offset:X}] # estimated
"""

    if exe.data_offset != 0:
        data_offset = exe.data_offset
        segments += f"""\
      - [0x{data_offset:X}, data, {data_offset:X}] # estimated
"""

    segments += f"""\
  - [0x{exe.size:X}]
"""

    out_file = Path(f"{cleaned_basename}.yaml")
    with out_file.open("w", encoding="utf-8", newline="\n") as f:
        print(f"Writing config to {out_file}")
        f.write(header)
        f.write(segments)

    # `file_presets` requires an initialized `opts`.
    # A simple way to do that is to simply load the yaml we just generated.
    conf.load([out_file])
    file_presets.write_all_files()


def do_elf(elf_path: Path, elf_bytes: bytes, objcopy: Optional[str]):
    elf = ps2elfinfo.Ps2Elf.get_info(elf_path, elf_bytes)
    if elf is None:
        log.error(f"Unsupported elf file '{elf_path}'")

    basename = elf_path.name.replace(" ", "")
    cleaned_basename = remove_invalid_path_characters(basename)

    rom_name = Path(f"{cleaned_basename}.rom")
    # Prefer the user objcopy
    if objcopy is None:
        objcopy = find_objcopy()
    objcopy_cmd = run_objcopy(objcopy, str(elf_path), str(rom_name))

    sha1 = hashlib.sha1(rom_name.read_bytes()).hexdigest()

    header = f"""\
# name: Your game name here!
sha1: {sha1}
options:
  basename: {basename}
  target_path: {rom_name}
  elf_path: build/{cleaned_basename}.elf
  base_path: .
  platform: ps2
  compiler: {elf.compiler}
"""

    if elf.gp is not None:
        header += f"""
  gp_value: 0x{elf.gp:08X}
"""
        if elf.ld_gp_expression is not None:
            header += f"  ld_gp_expression: {elf.ld_gp_expression}\n"
        else:
            header += "  # ld_gp_expression:\n"

    header += f"""
  # asm_path: asm
  # src_path: src
  # build_path: build

  ld_script_path: {cleaned_basename}.ld
  ld_dependencies: True
  ld_wildcard_sections: True
  ld_bss_contains_common: True

  create_asm_dependencies: True

  find_file_boundaries: False

  o_as_suffix: True

  symbol_addrs_path:
    - symbol_addrs.txt
  reloc_addrs_path:
    - reloc_addrs.txt

  # undefined_funcs_auto_path: undefined_funcs_auto.txt
  # undefined_syms_auto_path: undefined_syms_auto.txt

  extensions_path: tools/splat_ext

  string_encoding: ASCII
  data_string_encoding: ASCII
  rodata_string_guesser_level: 2
  data_string_guesser_level: 2

  named_regs_for_c_funcs: False
"""

    header += "\n  section_order:\n"
    for sect, is_valid in elf.elf_section_names:
        comment = "" if is_valid else "# "
        header += f"    {comment}- {sect}\n"

    header += "\n  auto_link_sections:\n"
    for sect, is_valid in elf.elf_section_names:
        comment = "" if is_valid else "# "
        if sect != ".text" and sect != ".vutext":
            header += f"    {comment}- {sect}\n"

    segments = "\nsegments:"
    for seg in elf.segs:
        segments += f"""
  - name: {seg.name}
    type: code
    start: 0x{seg.start:06X}
    vram: 0x{seg.vram:08X}
    bss_size: 0x{seg.bss_size:X}
    subalign: null
    subsegments:
"""
        for section in seg.sections:
            if section.is_nobits:
                segments += f"      - {{ type: {section.splat_segment_type}, vram: 0x{section.vram:08X}, name: {seg.name}/{section.vram:08X} }} # {section.name}\n"
            else:
                segments += f"      - [0x{section.start:06X}, {section.splat_segment_type}, {seg.name}/{section.start:06X}] # {section.name}\n"

    segments += f"""\
  - [0x{elf.size:X}]
"""

    out_file = Path(f"{cleaned_basename}.yaml")
    with out_file.open("w", encoding="utf-8", newline="\n") as f:
        print(f"Writing config to {out_file}")
        f.write(header)
        f.write(segments)

    # `file_presets` requires an initialized `opts`.
    # A simple way to do that is to simply load the yaml we just generated.
    conf.load([out_file])
    file_presets.write_all_files()

    # Write symbol_addrs.txt file
    symbol_addrs = []
    symbol_addrs.append(f"_start = 0x{elf.entrypoint:08X}; // type:func")
    if symbol_addrs:
        symbol_addrs.append("")
        with Path("symbol_addrs.txt").open("w", encoding="utf-8", newline="\n") as f:
            print("Writing symbol_addrs.txt")
            f.write(
                "// Visit https://github.com/ethteck/splat/wiki/Adding-Symbols for documentation about this file\n"
            )
            contents = "\n".join(symbol_addrs)
            f.write(contents)

    # Write other linker script
    linker_script = []
    linker_script.append("ENTRY(_start);")
    if linker_script:
        linker_script.append("")
        with Path("linker_script_extra.ld").open(
            "w",
            encoding="utf-8",
            newline="\n",
        ) as f:
            print("Writing linker_script_extra.ld")
            f.write(
                "/* Pass this file to the linker with the `-T linker_script_extra.ld` flag */\n"
            )
            contents = "\n".join(linker_script)
            f.write(contents)

    print()
    print(
        "The generated yaml does not use the actual ELF file as input, but instead it"
    )
    print(
        'uses a "rom" generated from said ELF, which contains the game code without any'
    )
    print("of the elf metadata.")
    print(
        'Use the following command to generate this "rom". It is recommended to include'
    )
    print("this command into your setup/configure script.")
    print("```")
    print(" ".join(objcopy_cmd))
    print("```")


def find_objcopy() -> str:
    # First we try to figure out if the user has objcopy on their pc, and under
    # which name.
    # We just try a bunch and hope for the best
    options = [
        "mips-linux-gnu-objcopy",
        "mipsel-linux-gnu-objcopy",
    ]

    for name in options:
        sub = subprocess.run([name, "--version"], capture_output=True)
        if sub.returncode == 0:
            return name

    msg = "Unable to find objcopy.\nI tried the following list of names:\n"
    for name in options:
        msg += f"  - {name}\n"
    msg += "\nTry to install one of those or use the `--objcopy` flag to pass the name to your own objcopy to me."
    log.error(msg)


def run_objcopy(objcopy_name: str, elf_path: str, rom: str) -> list[str]:
    cmd = [objcopy_name, "-O", "binary", "--gap-fill=0x00", elf_path, rom]
    print("Running:", " ".join(cmd))
    sub = subprocess.run(cmd)
    if sub.returncode != 0:
        log.error("Failed to run objcopy")
    return cmd


def add_arguments_to_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "file",
        help="Path to a .z64/.n64 ROM, PSX executable or PS2 ELF",
        type=Path,
    )
    parser.add_argument(
        "--objcopy",
        help="Path to an user-provided objcopy program. Only used when processing ELF files",
        type=str,
    )


def process_arguments(args: argparse.Namespace):
    main(args.file, args.objcopy)


script_description = "Create a splat config from an N64 ROM or PSX executable."


def add_subparser(subparser: argparse._SubParsersAction):
    parser = subparser.add_parser(
        "create_config", help=script_description, description=script_description
    )
    add_arguments_to_parser(parser)
    parser.set_defaults(func=process_arguments)


parser = argparse.ArgumentParser(description=script_description)
add_arguments_to_parser(parser)

if __name__ == "__main__":
    args = parser.parse_args()
    process_arguments(args)



from typing import Optional


from .codesubsegment import CommonSegCodeSubsegment


class CommonSegAsm(CommonSegCodeSubsegment):
    @staticmethod
    def is_text() -> bool:
        return True

    def get_section_flags(self) -> Optional[str]:
        return "ax"

    def scan(self, rom_bytes: bytes):
        if (
            self.rom_start is not None
            and self.rom_end is not None
            and self.rom_start != self.rom_end
        ):
            self.scan_code(rom_bytes)

    def split(self, rom_bytes: bytes):
        if self.rom_start == self.rom_end:
            return

        self.split_as_asm_file(self.out_path())



from ..common.header import CommonSegHeader


class PsxSegHeader(CommonSegHeader):
    # little endian so reverse words, TODO: use struct.unpack("<i",...) ?
    # breakdown from https://psx-spx.consoledev.net/cdromdrive/#filenameexe-general-purpose-executable
    def parse_header(self, rom_bytes):
        header_lines = []
        header_lines.append(".section .data\n")
        header_lines.append(
            self.get_line("ascii", rom_bytes[0x00:0x08], "Magic number")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x08:0x0C][::-1], ".text vram address")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x0C:0x10][::-1], ".data vram address")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x10:0x14][::-1], "Initial PC")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x14:0x18][::-1], "Initial $gp/r28")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x18:0x1C][::-1], ".text start")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x1C:0x20][::-1], ".text size")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x20:0x24][::-1], ".data start")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x24:0x28][::-1], ".data size")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x28:0x2C][::-1], ".bss start")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x2C:0x30][::-1], ".bss size")
        )
        header_lines.append(
            self.get_line(
                "word", rom_bytes[0x30:0x34][::-1], "Initial $sp/r29 & $fp/r30 base"
            )
        )
        header_lines.append(
            self.get_line(
                "word", rom_bytes[0x34:0x38][::-1], "Initial $sp/r29 & $fp/r30 offset"
            )
        )
        header_lines.append(self.get_line("word", rom_bytes[0x38:0x3C], "Reserved"))
        header_lines.append(self.get_line("word", rom_bytes[0x3C:0x40], "Reserved"))
        header_lines.append(self.get_line("word", rom_bytes[0x40:0x44], "Reserved"))
        header_lines.append(self.get_line("word", rom_bytes[0x44:0x48], "Reserved"))
        header_lines.append(self.get_line("word", rom_bytes[0x48:0x4C], "Reserved"))
        assert isinstance(self.rom_end, int)
        header_lines.append(
            self.get_line("ascii", rom_bytes[0x4C : self.rom_end], "Sony Inc")
        )

        header_lines.append("")

        return header_lines



from typing import Optional

from ..common.data import CommonSegData
from ...disassembler.disassembler_section import DisassemblerSection


class Ps2SegLit8(CommonSegData):
    """Segment that only contains double-precision floats"""

    def get_linker_section(self) -> str:
        return ".lit8"

    def get_section_flags(self) -> Optional[str]:
        return "wa"

    def configure_disassembler_section(
        self, disassembler_section: DisassemblerSection
    ) -> None:
        "Allows to configure the section before running the analysis on it"

        super().configure_disassembler_section(disassembler_section)

        section = disassembler_section.get_section()

        # Tell spimdisasm this section only contains doubles
        section.enableStringGuessing = False
        section.typeForOwnedSymbols = "f64"
        section.sizeForOwnedSymbols = 8



import n64img.image

from ...segtypes.n64.img import N64SegImg


class N64SegRgba32(N64SegImg):
    def __init__(self, *args, **kwargs):
        kwargs["img_cls"] = n64img.image.RGBA32
        super().__init__(*args, **kwargs)



#! /usr/bin/env python3

import argparse
import hashlib
import importlib
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from pathlib import Path

from collections import defaultdict, deque

from .. import __package_name__, __version__
from ..disassembler import disassembler_instance
from ..util import cache_handler, progress_bar, vram_classes, statistics, file_presets

from colorama import Style
from intervaltree import Interval, IntervalTree
import sys

from ..segtypes.linker_entry import (
    LinkerWriter,
    get_segment_vram_end_symbol_name,
)
from ..segtypes.segment import Segment
from ..segtypes.common.group import CommonSegGroup
from ..util import conf, log, options, palettes, symbols, relocs

linker_writer: LinkerWriter
config: Dict[str, Any]

segment_roms: IntervalTree = IntervalTree()
segment_rams: IntervalTree = IntervalTree()


def initialize_segments(config_segments: Union[dict, list]) -> List[Segment]:
    global segment_roms
    global segment_rams

    segment_roms = IntervalTree()
    segment_rams = IntervalTree()

    segments_by_name: Dict[str, Segment] = {}
    ret: List[Segment] = []

    # Cross segment pairing can be quite expensive, so we try to avoid it if the user haven't requested it.
    do_cross_segment_pairing = False

    last_rom_end = 0

    for i, seg_yaml in enumerate(config_segments):
        # end marker
        if isinstance(seg_yaml, list) and len(seg_yaml) == 1:
            continue

        seg_type = Segment.parse_segment_type(seg_yaml)

        segment_class = Segment.get_class_for_type(seg_type)

        this_start, is_auto_segment = Segment.parse_segment_start(seg_yaml)

        j = i + 1
        while j < len(config_segments):
            next_start, next_is_auto_segment = Segment.parse_segment_start(
                config_segments[j]
            )
            if next_start is not None:
                break
            j += 1
        if next_start is None:
            log.error(
                "Next segment address could not be found. Segments list must end with a rom end pos marker ([0x10000000])"
            )

        if segment_class.is_noload():
            # Pretend bss's rom address is after the last actual rom segment
            this_start = last_rom_end
            # and it has a rom size of zero
            next_start = last_rom_end

        segment: Segment = Segment.from_yaml(
            segment_class, seg_yaml, this_start, next_start, None
        )

        if segment.require_unique_name:
            if segment.name in segments_by_name:
                log.error(f"segment name '{segment.name}' is not unique")

            segments_by_name[segment.name] = segment

        if segment.pair_segment_name is not None:
            do_cross_segment_pairing = True

        ret.append(segment)
        if (
            isinstance(segment.rom_start, int)
            and isinstance(segment.rom_end, int)
            and segment.rom_start != segment.rom_end
        ):
            segment_roms.addi(segment.rom_start, segment.rom_end, segment)
        if (
            isinstance(segment.vram_start, int)
            and isinstance(segment.vram_end, int)
            and segment.vram_start != segment.vram_end
        ):
            segment_rams.addi(segment.vram_start, segment.vram_end, segment)

        if next_start is not None:
            last_rom_end = next_start

    for segment in ret:
        if segment.given_follows_vram:
            if segment.given_follows_vram not in segments_by_name:
                log.error(
                    f"segment '{segment.given_follows_vram}', the 'follows_vram' value for segment '{segment.name}', does not exist"
                )
            segment.given_vram_symbol = get_segment_vram_end_symbol_name(
                segments_by_name[segment.given_follows_vram]
            )

    if ret[-1].type == "pad":
        log.error(
            "Last segment in config cannot be a pad segment; see https://github.com/ethteck/splat/wiki/Segments#pad"
        )

    if do_cross_segment_pairing:
        # Do the cross segment pairing
        for seg in ret:
            if seg.pair_segment_name is not None:
                found = False
                for other_seg in ret:
                    if seg == other_seg:
                        continue
                    if seg.pair_segment_name == other_seg.name and isinstance(
                        other_seg, CommonSegGroup
                    ):
                        # We found the other segment specified by `pair_segment`

                        if other_seg.pair_segment_name is not None:
                            log.error(
                                f"Both segments '{seg.name}' and '{other_seg.name}' have a `pair_segment` value, when only at most one of the cross-paired segments can have this attribute."
                            )

                        # Not user error, hopefully...
                        assert seg.paired_segment is None, (
                            f"Somehow '{seg.name}' was already paired so something else? It is paired to '{seg.paired_segment.name}' instead of {other_seg.name}"
                        )
                        assert other_seg.paired_segment is None, (
                            f"Somehow '{other_seg.name}' was already paired so something else? It is paired to '{other_seg.paired_segment.name}' instead of {seg.name}"
                        )

                        found = True
                        # Pair them
                        seg.paired_segment = other_seg
                        other_seg.paired_segment = seg

                        if isinstance(seg, CommonSegGroup) and isinstance(
                            other_seg, CommonSegGroup
                        ):
                            # Pair the subsegments
                            seg.pair_subsegments_to_other_segment(other_seg)

                        break

                if not found:
                    log.error(
                        f"Segment '{seg.pair_segment_name}' not found.\n    It is referenced by segment '{seg.name}', because of the `pair_segment` attribute."
                    )

    return ret


def assign_symbols_to_segments():
    for symbol in symbols.all_symbols:
        if symbol.segment:
            continue

        if symbol.rom:
            cands: Set[Interval] = segment_roms[symbol.rom]
            if len(cands) > 1:
                log.error("multiple segments rom overlap symbol", symbol)
            elif len(cands) == 0:
                log.error("no segment rom overlaps symbol", symbol)
            else:
                cand: Interval = cands.pop()
                seg: Segment = cand.data
                seg.add_symbol(symbol)
        else:
            cands = segment_rams[symbol.vram_start]
            segs: List[Segment] = [cand.data for cand in cands]
            for seg in segs:
                if not seg.get_exclusive_ram_id():
                    seg.add_symbol(symbol)


def brief_seg_name(seg: Segment, limit: int, ellipsis="…") -> str:
    s = seg.name.strip()
    if len(s) > limit:
        return s[:limit].strip() + ellipsis
    return s


# Return a mapping of vram classes to segments that need to be part of their vram symbol's calculation
def calc_segment_dependences(
    all_segments: List[Segment],
) -> Dict[vram_classes.VramClass, List[Segment]]:
    # Map vram class names to segments that have that vram class
    vram_class_to_segments: Dict[str, List[Segment]] = {}
    for seg in all_segments:
        if seg.vram_class is not None:
            if seg.vram_class.name not in vram_class_to_segments:
                vram_class_to_segments[seg.vram_class.name] = []
            vram_class_to_segments[seg.vram_class.name].append(seg)

    # Map vram class names to segments that the vram class follows
    vram_class_to_follows_segments: Dict[vram_classes.VramClass, List[Segment]] = {}
    for vram_class in vram_classes._vram_classes.values():
        if vram_class.follows_classes:
            vram_class_to_follows_segments[vram_class] = []

            for follows_class in vram_class.follows_classes:
                if follows_class in vram_class_to_segments:
                    vram_class_to_follows_segments[vram_class] += (
                        vram_class_to_segments[follows_class]
                    )
    return vram_class_to_follows_segments


def sort_segments_by_vram_class_dependency(
    all_segments: List[Segment],
) -> List[Segment]:
    # map all "_VRAM_END" strings to segments
    end_sym_to_seg: Dict[str, Segment] = {}
    for seg in all_segments:
        end_sym_to_seg[get_segment_vram_end_symbol_name(seg)] = seg

    # build dependency graph: A -> B means "A must come before B"
    graph: Dict[Segment, List[Segment]] = defaultdict(list)
    indeg: Dict[Segment, int] = {seg: 0 for seg in all_segments}

    for seg in all_segments:
        sym = seg.vram_symbol
        if sym is None:
            continue
        dep = end_sym_to_seg.get(sym)
        if dep is None or dep is seg:
            continue
        graph[dep].append(seg)
        indeg[seg] += 1

    # stable topo sort with queue seeded in original order
    q = deque([seg for seg in all_segments if indeg[seg] == 0])
    out: List[Segment] = []

    while q:
        n = q.popleft()
        out.append(n)
        for m in graph.get(n, []):
            indeg[m] -= 1
            if indeg[m] == 0:
                q.append(m)

    assert len(out) == len(all_segments), (
        "Encountered cyclic dependency when reordering segments by vram class."
    )

    return out


def read_target_binary() -> bytes:
    rom_bytes = options.opts.target_path.read_bytes()

    if "sha1" in config:
        sha1 = hashlib.sha1(rom_bytes).hexdigest()
        e_sha1 = config["sha1"].lower()
        if e_sha1 != sha1:
            log.error(f"sha1 mismatch: expected {e_sha1}, was {sha1}")
    else:
        log.write("Warning: no sha1 in config")

    return rom_bytes


def initialize_platform(rom_bytes: bytes):
    platform_module = importlib.import_module(
        f"{__package_name__}.platforms.{options.opts.platform}"
    )
    platform_init = getattr(platform_module, "init")
    platform_init(rom_bytes)

    return platform_module


def initialize_all_symbols(all_segments: List[Segment]):
    # Load and process symbols
    symbols.initialize(all_segments)
    relocs.initialize()

    # Assign symbols to segments
    assign_symbols_to_segments()

    if options.opts.is_mode_active("code"):
        symbols.initialize_spim_context(all_segments)
        relocs.initialize_spim_context()


def do_scan(
    all_segments: List[Segment],
    rom_bytes: bytes,
    stats: statistics.Statistics,
    cache: cache_handler.Cache,
):
    processed_segments: List[Segment] = []

    scan_bar = progress_bar.get_progress_bar(all_segments)
    for segment in scan_bar:
        assert isinstance(segment, Segment)
        scan_bar.set_description(f"Scanning {brief_seg_name(segment, 20)}")

        for ty, sub_stats in segment.statistics.items():
            stats.add_size(ty, sub_stats.size)

        if segment.should_scan():
            # Check cache but don't write anything
            if cache.check_cache_hit(segment, False):
                continue

            segment.did_run = True
            segment.scan(rom_bytes)

            processed_segments.append(segment)

            for ty, sub_stats in segment.statistics.items():
                stats.count_split(ty, sub_stats.count)

    symbols.mark_c_funcs_as_defined()
    return processed_segments


def do_split(
    all_segments: List[Segment],
    rom_bytes: bytes,
    stats: statistics.Statistics,
    cache: cache_handler.Cache,
):
    split_bar = progress_bar.get_progress_bar(all_segments)
    for segment in split_bar:
        assert isinstance(segment, Segment)
        split_bar.set_description(f"Splitting {brief_seg_name(segment, 20)}")

        if cache.check_cache_hit(segment, True):
            for ty, sub_stats in segment.statistics.items():
                stats.count_cached(ty, sub_stats.count)
            continue

        if segment.should_split():
            segment_bytes = rom_bytes
            if segment.file_path:
                segment_bytes = segment.file_path.read_bytes()
            segment.split(segment_bytes)


def write_linker_script(all_segments: List[Segment]) -> LinkerWriter:
    if options.opts.ld_sort_segments_by_vram_class_dependency:
        all_segments = sort_segments_by_vram_class_dependency(all_segments)

    vram_class_dependencies = calc_segment_dependences(all_segments)
    vram_classes_to_search = set(vram_class_dependencies.keys())

    max_vram_end_insertion_points: Dict[Segment, List[Tuple[str, List[Segment]]]] = {}
    for seg in reversed(all_segments):
        if seg.vram_class in vram_classes_to_search:
            assert seg.vram_class.vram_symbol is not None
            if seg not in max_vram_end_insertion_points:
                max_vram_end_insertion_points[seg] = []
            max_vram_end_insertion_points[seg].append(
                (
                    seg.vram_class.vram_symbol,
                    vram_class_dependencies[seg.vram_class],
                )
            )
            vram_classes_to_search.remove(seg.vram_class)

    linker_writer = LinkerWriter()
    linker_bar = progress_bar.get_progress_bar(all_segments)

    # Check options are valid
    partial_linking = options.opts.ld_partial_linking
    partial_scripts_path = options.opts.ld_partial_scripts_path
    segments_path = options.opts.ld_partial_build_segments_path
    if partial_linking:
        if partial_scripts_path is None:
            log.error(
                "Partial linking is enabled but `ld_partial_scripts_path` has not been set"
            )
        if options.opts.ld_partial_build_segments_path is None:
            log.error(
                "Partial linking is enabled but `ld_partial_build_segments_path` has not been set"
            )

    for segment in linker_bar:
        assert isinstance(segment, Segment)
        linker_bar.set_description(f"Linker script {brief_seg_name(segment, 20)}")
        max_vram_syms = max_vram_end_insertion_points.get(segment, [])

        if options.opts.ld_partial_linking:
            linker_writer.add_referenced_partial_segment(segment, max_vram_syms)

            # Create linker script for segment
            sub_linker_writer = LinkerWriter(is_partial=True)
            sub_linker_writer.add_partial_segment(segment)

            assert partial_scripts_path is not None
            assert segments_path is not None

            seg_name = segment.get_cname()

            sub_linker_writer.save_linker_script(
                partial_scripts_path / f"{seg_name}.ld"
            )
            if options.opts.ld_dependencies:
                sub_linker_writer.save_dependencies_file(
                    partial_scripts_path / f"{seg_name}.d",
                    segments_path / f"{seg_name}.o",
                )
        else:
            linker_writer.add(segment, max_vram_syms)

    linker_writer.save_linker_script(options.opts.ld_script_path)
    linker_writer.save_symbol_header()

    if options.opts.ld_dependencies:
        elf_path = options.opts.elf_path
        if elf_path is None:
            log.error(
                "Generation of dependency file for linker script requested but `elf_path` was not provided in the yaml options"
            )
        linker_writer.save_dependencies_file(
            options.opts.ld_script_path.with_suffix(".d"), elf_path
        )

    return linker_writer


def write_ld_dependencies(linker_writer: LinkerWriter):
    if options.opts.ld_dependencies:
        elf_path = options.opts.elf_path
        if elf_path is None:
            log.error(
                "Generation of dependency file for linker script requested but `elf_path` was not provided in the yaml options"
            )
        linker_writer.save_dependencies_file(
            options.opts.ld_script_path.with_suffix(".d"), elf_path
        )


def write_elf_sections_file(all_segments: List[Segment]):
    # write elf_sections.txt - this only lists the generated sections in the elf, not subsections
    # that the elf combines into one section
    if options.opts.elf_section_list_path:
        section_list = ""
        for segment in all_segments:
            section_list += "." + segment.get_cname() + "\n"
        options.opts.elf_section_list_path.parent.mkdir(parents=True, exist_ok=True)
        with options.opts.elf_section_list_path.open(
            "w",
            encoding="utf-8",
            newline="\n",
        ) as f:
            f.write(section_list)


def write_undefined_auto(to_write: List[symbols.Symbol], file_path: Path):
    file_path.parent.mkdir(parents=True, exist_ok=True)
    with file_path.open("w", encoding="utf-8", newline="\n") as f:
        for symbol in to_write:
            f.write(f"{symbol.name} = 0x{symbol.vram_start:X};\n")


def write_undefined_funcs_auto():
    if options.opts.create_undefined_funcs_auto:
        to_write = [
            s
            for s in symbols.all_symbols
            if s.referenced and not s.defined and s.type == "func"
        ]
        to_write.sort(key=lambda x: x.vram_start)

        write_undefined_auto(to_write, options.opts.undefined_funcs_auto_path)


def write_undefined_syms_auto():
    if options.opts.create_undefined_syms_auto:
        to_write = [
            s
            for s in symbols.all_symbols
            if s.referenced
            and not s.defined
            and s.type not in {"func", "label", "jtbl_label"}
        ]
        to_write.sort(key=lambda x: x.vram_start)

        write_undefined_auto(to_write, options.opts.undefined_syms_auto_path)


def print_segment_warnings(all_segments: List[Segment]):
    for segment in all_segments:
        if len(segment.warnings) > 0:
            log.write(
                f"{Style.DIM}0x{segment.rom_start:06X}{Style.RESET_ALL} {segment.type} {Style.BRIGHT}{segment.name}{Style.RESET_ALL}:"
            )

            for warn in segment.warnings:
                log.write("warning: " + warn, status="warn")

            log.write("")  # empty line


def dump_symbols() -> None:
    if not options.opts.dump_symbols:
        return

    splat_hidden_folder = options.opts.base_path / ".splat"
    splat_hidden_folder.mkdir(parents=True, exist_ok=True)

    with open(splat_hidden_folder / "splat_symbols.csv", "w", encoding="utf-8") as f:
        f.write(
            "vram_start,given_name,name,type,given_size,size,rom,defined,user_declared,referenced,extract\n"
        )
        for s in sorted(symbols.all_symbols, key=lambda x: x.vram_start):
            f.write(f"{s.vram_start:X},{s.given_name},{s.name},{s.type},")
            if s.given_size is not None:
                f.write(f"0x{s.given_size:X},")
            else:
                f.write("None,")
            f.write(f"{s.size},")
            if s.rom is not None:
                f.write(f"0x{s.rom:X},")
            else:
                f.write("None,")
            f.write(f"{s.defined},{s.user_declared},{s.referenced},{s.extract}\n")

    symbols.spim_context.saveContextToFile(splat_hidden_folder / "spim_context.csv")


def main(
    config_path: List[Path],
    modes: Optional[List[str]],
    verbose: bool,
    use_cache: bool = True,
    skip_version_check: bool = False,
    stdout_only: bool = False,
    disassemble_all: bool = False,
    make_full_disasm_for_code=False,
):
    if stdout_only:
        log.write("--stdout-only flag is deprecated", status="warn")
    progress_bar.out_file = sys.stdout

    # Load config
    global config
    config = conf.load(
        config_path,
        modes,
        verbose,
        disassemble_all,
        make_full_disasm_for_code,
    )

    disassembler_instance.create_disassembler_instance(skip_version_check, __version__)

    rom_bytes = read_target_binary()

    # Create main output dir
    options.opts.base_path.mkdir(parents=True, exist_ok=True)

    stats = statistics.Statistics()

    cache = cache_handler.Cache(config, use_cache, verbose)

    if not options.opts.is_unsupported_platform:
        initialize_platform(rom_bytes)

    # Initialize segments
    all_segments = initialize_segments(config["segments"])

    initialize_all_symbols(all_segments)

    # Resolve raster/palette siblings
    if options.opts.is_mode_active("img"):
        palettes.initialize(all_segments)

    # Scan
    do_scan(all_segments, rom_bytes, stats, cache)

    # Split
    do_split(all_segments, rom_bytes, stats, cache)

    if options.opts.is_mode_active(
        "ld"
    ):  # TODO move this to platform initialization when it gets implemented
        global linker_writer
        linker_writer = write_linker_script(all_segments)
        write_ld_dependencies(linker_writer)
        write_elf_sections_file(all_segments)

    # Write undefined_funcs_auto.txt
    write_undefined_funcs_auto()

    # write undefined_syms_auto.txt
    write_undefined_syms_auto()

    # print warnings during split
    print_segment_warnings(all_segments)

    # Statistics
    stats.print_statistics(len(rom_bytes))

    # Save cache
    cache.save(verbose)

    if options.opts.is_mode_active("code"):
        dump_symbols()

        file_presets.write_all_files()


def add_arguments_to_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "config",
        help="path to a compatible config .yaml file",
        nargs="+",
        type=Path,
    )
    parser.add_argument("--modes", nargs="+", default=["all"])
    parser.add_argument("--verbose", action="store_true", help="Enable debug logging")
    parser.add_argument(
        "--use-cache", action="store_true", help="Only split changed segments in config"
    )
    parser.add_argument(
        "--skip-version-check",
        action="store_true",
        help="Skips the disassembler's version check",
    )
    parser.add_argument(
        "--stdout-only",
        help="Print all output to stdout (deprecated)",
        action="store_true",
    )
    parser.add_argument(
        "--disassemble-all",
        help="Disasemble matched functions and migrated data",
        action="store_true",
    )
    parser.add_argument(
        "--make-full-disasm-for-code",
        help="Emit a full `.s` file for each `c`/`cpp` segment besides the generated `nonmatchings` individual functions",
        action="store_true",
    )


def process_arguments(args: argparse.Namespace):
    main(
        args.config,
        args.modes,
        args.verbose,
        args.use_cache,
        args.skip_version_check,
        args.stdout_only,
        args.disassemble_all,
        args.make_full_disasm_for_code,
    )


script_description = "Split a rom given a rom, a config, and output directory"


def add_subparser(subparser: argparse._SubParsersAction):
    parser = subparser.add_parser(
        "split", help=script_description, description=script_description
    )
    add_arguments_to_parser(parser)
    parser.set_defaults(func=process_arguments)


parser = argparse.ArgumentParser(description=script_description)
add_arguments_to_parser(parser)

if __name__ == "__main__":
    args = parser.parse_args()
    process_arguments(args)



from . import disassembler
import spimdisasm
import rabbitizer
from ..util import log, compiler, options
from typing import Set


class SpimdisasmDisassembler(disassembler.Disassembler):
    # This value should be kept in sync with the version listed on requirements.txt and pyproject.toml
    SPIMDISASM_MIN = (1, 40, 0)

    def configure(self):
        # Configure spimdisasm
        spimdisasm.common.GlobalConfig.PRODUCE_SYMBOLS_PLUS_OFFSET = True
        spimdisasm.common.GlobalConfig.TRUST_USER_FUNCTIONS = True
        spimdisasm.common.GlobalConfig.TRUST_JAL_FUNCTIONS = True
        spimdisasm.common.GlobalConfig.GLABEL_ASM_COUNT = False

        if options.opts.rom_address_padding:
            spimdisasm.common.GlobalConfig.ASM_COMMENT_OFFSET_WIDTH = 6
        else:
            spimdisasm.common.GlobalConfig.ASM_COMMENT_OFFSET_WIDTH = 0

        # spimdisasm is not performing any analyzis on non-text sections so enabling this options is pointless
        spimdisasm.common.GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_SECTION_TYPE = False
        spimdisasm.common.GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_DATA_TYPE = False

        spimdisasm.common.GlobalConfig.SYMBOL_FINDER_FILTERED_ADDRESSES_AS_HILO = False

        if options.opts.rodata_string_guesser_level is not None:
            spimdisasm.common.GlobalConfig.RODATA_STRING_GUESSER_LEVEL = (
                options.opts.rodata_string_guesser_level
            )

        if options.opts.data_string_guesser_level is not None:
            spimdisasm.common.GlobalConfig.DATA_STRING_GUESSER_LEVEL = (
                options.opts.data_string_guesser_level
            )

        rabbitizer.config.regNames_userFpcCsr = False
        rabbitizer.config.regNames_vr4300Cop0NamedRegisters = False

        rabbitizer.config.misc_opcodeLJust = options.opts.mnemonic_ljust - 1

        rabbitizer.config.regNames_gprAbiNames = rabbitizer.Abi.fromStr(
            options.opts.mips_abi_gpr
        )
        rabbitizer.config.regNames_fprAbiNames = rabbitizer.Abi.fromStr(
            options.opts.mips_abi_float_regs
        )

        if options.opts.endianness == "big":
            spimdisasm.common.GlobalConfig.ENDIAN = spimdisasm.common.InputEndian.BIG
        else:
            spimdisasm.common.GlobalConfig.ENDIAN = spimdisasm.common.InputEndian.LITTLE

        rabbitizer.config.pseudos_pseudoMove = False

        selected_compiler = options.opts.compiler
        spimdisasm_compiler = spimdisasm.common.Compiler.fromStr(selected_compiler.name)
        if spimdisasm_compiler is None:
            log.write(
                f"Unsupported selected compiler for spimdisasm: {selected_compiler.name}",
                status="error",
            )
            log.error(
                f"The following options are supported: {list(spimdisasm.common.compilerOptions.keys())}"
            )
        spimdisasm.common.GlobalConfig.COMPILER = spimdisasm_compiler
        if selected_compiler == compiler.SN64:
            rabbitizer.config.regNames_namedRegisters = False
            rabbitizer.config.toolchainTweaks_sn64DivFix = True
            spimdisasm.common.GlobalConfig.ASM_COMMENT = False
            spimdisasm.common.GlobalConfig.SYMBOL_FINDER_FILTERED_ADDRESSES_AS_HILO = (
                False
            )
        if options.opts.align_on_branch_labels:
            spimdisasm.common.GlobalConfig.ASM_EMIT_ALIGN_BRANCH_LABELS = True
        rabbitizer.config.toolchainTweaks_treatJAsUnconditionalBranch = (
            selected_compiler.j_as_branch
        )

        spimdisasm.common.GlobalConfig.DETECT_REDUNDANT_FUNCTION_END = (
            options.opts.detect_redundant_function_end
        )

        spimdisasm.common.GlobalConfig.GP_VALUE = options.opts.gp

        spimdisasm.common.GlobalConfig.ASM_TEXT_LABEL = options.opts.asm_function_macro
        spimdisasm.common.GlobalConfig.ASM_TEXT_ALT_LABEL = (
            options.opts.asm_function_alt_macro
        )
        spimdisasm.common.GlobalConfig.ASM_JTBL_LABEL = (
            options.opts.asm_jtbl_label_macro
        )
        spimdisasm.common.GlobalConfig.ASM_DATA_LABEL = options.opts.asm_data_macro
        spimdisasm.common.GlobalConfig.ASM_TEXT_END_LABEL = options.opts.asm_end_label
        spimdisasm.common.GlobalConfig.ASM_DATA_END_LABEL = (
            options.opts.asm_data_end_label
        )
        spimdisasm.common.GlobalConfig.ASM_EHTBL_LABEL = (
            options.opts.asm_ehtable_label_macro
        )
        spimdisasm.common.GlobalConfig.ASM_NM_LABEL = (
            options.opts.asm_nonmatching_label_macro
        )

        if options.opts.asm_emit_size_directive is not None:
            spimdisasm.common.GlobalConfig.ASM_EMIT_SIZE_DIRECTIVE = (
                options.opts.asm_emit_size_directive
            )

        if spimdisasm.common.GlobalConfig.ASM_TEXT_LABEL == ".globl":
            spimdisasm.common.GlobalConfig.ASM_TEXT_ENT_LABEL = ".ent"
            spimdisasm.common.GlobalConfig.ASM_TEXT_FUNC_AS_LABEL = True

        if spimdisasm.common.GlobalConfig.ASM_DATA_LABEL == ".globl":
            spimdisasm.common.GlobalConfig.ASM_DATA_SYM_AS_LABEL = True

        spimdisasm.common.GlobalConfig.LINE_ENDS = options.opts.c_newline

        spimdisasm.common.GlobalConfig.ALLOW_ALL_ADDENDS_ON_DATA = (
            options.opts.allow_data_addends
        )

        spimdisasm.common.GlobalConfig.DISASSEMBLE_UNKNOWN_INSTRUCTIONS = (
            options.opts.disasm_unknown
        )

    def check_version(self, skip_version_check: bool, splat_version: str):
        if not skip_version_check and spimdisasm.__version_info__ < self.SPIMDISASM_MIN:
            log.error(
                f"splat {splat_version} requires as minimum spimdisasm {self.SPIMDISASM_MIN}, but the installed version is {spimdisasm.__version_info__}"
            )

        log.write(
            f"splat {splat_version} (powered by spimdisasm {spimdisasm.__version__})"
        )

    def known_types(self) -> Set[str]:
        return spimdisasm.common.gKnownTypes



#!/usr/bin/env python3

import argparse

import splat


def splat_main():
    parser = argparse.ArgumentParser(
        description="A binary splitting tool to assist with decompilation and modding projects",
        prog="splat",
    )

    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {splat.__version__}"
    )

    subparsers = parser.add_subparsers(
        description="action", help="The CLI utility to run", required=True
    )

    splat.scripts.split.add_subparser(subparsers)
    splat.scripts.create_config.add_subparser(subparsers)
    splat.scripts.capy.add_subparser(subparsers)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    splat_main()



import collections
import dataclasses
import importlib
import importlib.util
from pathlib import Path

from typing import Dict, List, Optional, Set, Type, TYPE_CHECKING, Union, Tuple

from intervaltree import Interval, IntervalTree
from ..util import vram_classes

from ..util.vram_classes import VramClass
from ..util import log, options, symbols
from ..util.symbols import Symbol, to_cname

from .. import __package_name__

# circular import
if TYPE_CHECKING:
    from ..segtypes.linker_entry import LinkerEntry


def parse_segment_vram(segment: Union[dict, list]) -> Optional[int]:
    if isinstance(segment, dict) and "vram" in segment:
        return int(segment["vram"])
    else:
        return None


def parse_segment_vram_symbol(segment: Union[dict, list]) -> Optional[str]:
    if isinstance(segment, dict) and "vram_symbol" in segment:
        return str(segment["vram_symbol"])
    else:
        return None


def parse_segment_vram_class(segment: Union[dict, list]) -> Optional[VramClass]:
    if isinstance(segment, dict) and "vram_class" in segment:
        return vram_classes.resolve(segment["vram_class"])
    return None


def parse_segment_follows_vram(segment: Union[dict, list]) -> Optional[str]:
    if isinstance(segment, dict):
        return segment.get("follows_vram", None)
    return None


def parse_segment_align(segment: Union[dict, list]) -> Optional[int]:
    if isinstance(segment, dict) and "align" in segment:
        return int(segment["align"])
    return None


def parse_segment_subalign(segment: Union[dict, list]) -> Optional[int]:
    default = options.opts.subalign
    if isinstance(segment, dict):
        subalign = segment.get("subalign", default)
        if subalign is not None:
            subalign = int(subalign)
        return subalign
    return default


def parse_segment_section_order(segment: Union[dict, list]) -> List[str]:
    default = options.opts.section_order
    if isinstance(segment, dict):
        return segment.get("section_order", default)
    return default


SegmentType = str


@dataclasses.dataclass
class SegmentStatisticsInfo:
    size: int
    count: int

    def merge(self, other: "SegmentStatisticsInfo") -> "SegmentStatisticsInfo":
        return SegmentStatisticsInfo(
            size=self.size + other.size, count=self.count + other.count
        )


SegmentStatistics = dict[SegmentType, SegmentStatisticsInfo]


def empty_statistics() -> SegmentStatistics:
    return collections.defaultdict(lambda: SegmentStatisticsInfo(size=0, count=0))


class Segment:
    require_unique_name = True

    @staticmethod
    def get_class_for_type(seg_type) -> Type["Segment"]:
        # so .data loads SegData, for example
        if seg_type.startswith("."):
            seg_type = seg_type[1:]

        if options.opts.allow_segment_overrides:
            segment_class = Segment.get_extension_segment_class(seg_type)
            if segment_class is None:
                segment_class = Segment.get_base_segment_class(seg_type)
        else:
            segment_class = Segment.get_base_segment_class(seg_type)
            if segment_class is None:
                # Look in extensions
                segment_class = Segment.get_extension_segment_class(seg_type)

        if segment_class is None:
            log.error(
                f"could not load segment type '{seg_type}'\n(hint: confirm your extension directory is configured correctly)"
            )

        return segment_class

    @staticmethod
    def get_base_segment_class(seg_type):
        platform = options.opts.platform
        is_platform_seg = False

        # heirarchy is platform -> common -> fail
        try:
            segmodule = importlib.import_module(
                f".segtypes.{platform}.{seg_type}", package=__package_name__
            )
            is_platform_seg = True
        except ModuleNotFoundError:
            try:
                segmodule = importlib.import_module(
                    f".segtypes.common.{seg_type}", package=__package_name__
                )
            except ModuleNotFoundError:
                return None

        seg_prefix = platform.capitalize() if is_platform_seg else "Common"
        return getattr(segmodule, f"{seg_prefix}Seg{seg_type.capitalize()}")

    @staticmethod
    def get_extension_segment_class(seg_type):
        platform = options.opts.platform

        ext_path = options.opts.extensions_path
        if not ext_path:
            log.error(
                f"could not load presumed extended segment type '{seg_type}' because no extensions path is configured"
            )
        assert ext_path is not None

        try:
            ext_spec = importlib.util.spec_from_file_location(
                f"{__package_name__}.segtypes.{platform}.{seg_type}",
                ext_path / f"{seg_type}.py",
            )
            assert ext_spec is not None
            ext_mod = importlib.util.module_from_spec(ext_spec)
            assert ext_spec.loader is not None
            ext_spec.loader.exec_module(ext_mod)
        except Exception:
            return None

        return getattr(
            ext_mod, f"{platform.upper()}Seg{seg_type[0].upper()}{seg_type[1:]}"
        )

    @staticmethod
    def parse_segment_start(segment: Union[dict, list]) -> Tuple[Optional[int], bool]:
        """
        Parses the rom start address of a given segment.

        Returns a two-tuple containing:
        - The rom start address of the segment, if any.
        - `True` if the user explicitly specified `auto` as the start address.
          Note this will be `False` if user specified an actual number or did not specify anything at all (in the dict notation).
          Not specifying a explicit `start` is useful for `bss`/`sbss` segments, since they do not have a real rom address.
        """

        if isinstance(segment, dict):
            s = segment.get("start", None)
        else:
            s = segment[0]

        if s is None:
            return None, False
        if s == "auto":
            return None, True
        else:
            return int(s), False

    @staticmethod
    def parse_segment_type(segment: Union[dict, list]) -> str:
        if isinstance(segment, dict):
            return str(segment["type"])
        else:
            return str(segment[1])

    @staticmethod
    def parse_segment_name(cls, rom_start, segment: Union[dict, list]) -> str:
        if isinstance(segment, dict) and "name" in segment:
            return str(segment["name"])
        elif isinstance(segment, dict) and "dir" in segment:
            return str(segment["dir"])
        elif isinstance(segment, list) and len(segment) >= 3:
            return str(segment[2])
        else:
            return str(cls.get_default_name(rom_start))

    @staticmethod
    def parse_segment_symbol_name_format(segment: Union[dict, list]) -> str:
        if isinstance(segment, dict) and "symbol_name_format" in segment:
            return str(segment["symbol_name_format"])
        else:
            return options.opts.symbol_name_format

    @staticmethod
    def parse_segment_symbol_name_format_no_rom(segment: Union[dict, list]) -> str:
        if isinstance(segment, dict) and "symbol_name_format_no_rom" in segment:
            return str(segment["symbol_name_format_no_rom"])
        else:
            return options.opts.symbol_name_format_no_rom

    @staticmethod
    def parse_segment_file_path(segment: Union[dict, list]) -> Optional[Path]:
        if isinstance(segment, dict) and "path" in segment:
            return Path(segment["path"])
        return None

    @staticmethod
    def parse_segment_bss_contains_common(
        segment: Union[dict, list], default: bool
    ) -> bool:
        if isinstance(segment, dict) and "bss_contains_common" in segment:
            return bool(segment["bss_contains_common"])
        return default

    @staticmethod
    def parse_linker_section_order(yaml: Union[dict, list]) -> Optional[str]:
        if isinstance(yaml, dict) and "linker_section_order" in yaml:
            return str(yaml["linker_section_order"])
        return None

    @staticmethod
    def parse_linker_section(yaml: Union[dict, list]) -> Optional[str]:
        if isinstance(yaml, dict) and "linker_section" in yaml:
            return str(yaml["linker_section"])
        return None

    @staticmethod
    def parse_ld_fill_value(
        yaml: Union[dict, list], default: Optional[int]
    ) -> Optional[int]:
        if isinstance(yaml, dict) and "ld_fill_value" in yaml:
            return yaml["ld_fill_value"]
        return default

    @staticmethod
    def parse_ld_align_segment_start(yaml: Union[dict, list]) -> Optional[int]:
        if isinstance(yaml, dict) and "ld_align_segment_start" in yaml:
            return yaml["ld_align_segment_start"]
        return options.opts.ld_align_segment_start

    @staticmethod
    def parse_suggestion_rodata_section_start(
        yaml: Union[dict, list],
    ) -> Optional[bool]:
        if isinstance(yaml, dict):
            suggestion_rodata_section_start = yaml.get(
                "suggestion_rodata_section_start"
            )
            if suggestion_rodata_section_start is not None:
                assert isinstance(suggestion_rodata_section_start, bool)
                return suggestion_rodata_section_start
        return None

    @staticmethod
    def parse_pair_segment(yaml: Union[dict, list]) -> Optional[str]:
        if isinstance(yaml, dict) and "pair_segment" in yaml:
            return yaml["pair_segment"]
        return None

    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
    ):
        self.rom_start = rom_start
        self.rom_end = rom_end
        self.type = type
        self.name = name
        self.vram_start: Optional[int] = vram_start

        self.align: Optional[int] = None
        self.given_subalign: Optional[int] = options.opts.subalign
        self.exclusive_ram_id: Optional[str] = None
        self.given_dir: Path = Path()

        # Default to global options.
        self.given_find_file_boundaries: Optional[bool] = None

        # Symbols known to be in this segment
        self.given_seg_symbols: Dict[int, List[Symbol]] = {}

        # Ranges for faster symbol lookup
        self.symbol_ranges_ram: IntervalTree = IntervalTree()
        self.symbol_ranges_rom: IntervalTree = IntervalTree()

        self.given_section_order: List[str] = options.opts.section_order

        self.vram_class: Optional[VramClass] = None
        self.given_follows_vram: Optional[str] = None
        self.given_vram_symbol: Optional[str] = None

        self.given_symbol_name_format: str = options.opts.symbol_name_format
        self.given_symbol_name_format_no_rom: str = (
            options.opts.symbol_name_format_no_rom
        )

        self.parent: Optional[Segment] = None
        self.sibling: Optional[Segment] = None
        self.siblings: Dict[str, Segment] = {}
        self.pair_segment_name: Optional[str] = self.parse_pair_segment(yaml)
        self.paired_segment: Optional[Segment] = None

        self.file_path: Optional[Path] = None

        self.args: List[str] = args
        self.yaml = yaml

        self.extract: bool = True
        self.has_linker_entry: bool = True
        if self.rom_start is None:
            self.extract = False
        elif self.type.startswith("."):
            self.extract = False

        self.warnings: List[str] = []
        self.did_run = False
        self.bss_contains_common = Segment.parse_segment_bss_contains_common(
            yaml, options.opts.ld_bss_contains_common
        )

        # For segments which are not in the usual VRAM segment space, like N64's IPL3 which lives in 0xA4...
        self.special_vram_segment: bool = False

        self.linker_section_order: Optional[str] = self.parse_linker_section_order(yaml)
        self.linker_section: Optional[str] = self.parse_linker_section(yaml)

        # If not defined on the segment then default to the global option
        self.ld_fill_value: Optional[int] = self.parse_ld_fill_value(
            yaml, options.opts.ld_fill_value
        )

        self.ld_align_segment_start: Optional[int] = self.parse_ld_align_segment_start(
            yaml
        )

        # True if this segment was generated based on auto_link_sections
        self.is_generated: bool = False

        self.given_suggestion_rodata_section_start: Optional[bool] = (
            self.parse_suggestion_rodata_section_start(yaml)
        )

        # Is an automatic segment, generated automatically or declared on the yaml by the user
        self.is_auto_segment: bool = False

        self.index_within_group: Optional[int] = None

        if self.rom_start is not None and self.rom_end is not None:
            if self.rom_start > self.rom_end:
                log.error(
                    f"Error: segments out of order - ({self.name} starts at 0x{self.rom_start:X}, but next segment starts at 0x{self.rom_end:X})"
                )

    @staticmethod
    def from_yaml(
        cls: Type["Segment"],
        yaml: Union[dict, list],
        rom_start: Optional[int],
        rom_end: Optional[int],
        parent: Optional["Segment"],
        vram=None,
    ):
        type = Segment.parse_segment_type(yaml)
        name = Segment.parse_segment_name(cls, rom_start, yaml)

        vram_class = parse_segment_vram_class(yaml)

        if vram is not None:
            vram_start = vram
        elif vram_class:
            vram_start = vram_class.vram
        else:
            vram_start = parse_segment_vram(yaml)

        args: List[str] = [] if isinstance(yaml, dict) else yaml[3:]

        ret = cls(
            rom_start=rom_start,
            rom_end=rom_end,
            type=type,
            name=name,
            vram_start=vram_start,
            args=args,
            yaml=yaml,
        )
        if parent is not None:
            if "subalign" in yaml:
                log.error(
                    f"Non top-level segment '{name}' (rom address 0x{rom_start:X}) specified a `subalign`. `subalign` is valid only for top-level segments"
                )
            if "ld_fill_value" in yaml:
                log.error(
                    f"Non top-level segment '{name}' (rom address 0x{rom_start:X}) specified a `ld_fill_value`. `ld_fill_value` is valid only for top-level segments"
                )

        ret.parent = parent

        # Import here to avoid circular imports
        from .common.code import CommonSegCode
        from .common.bss import CommonSegBss

        if (
            options.opts.ld_bss_is_noload
            and isinstance(ret, CommonSegBss)
            and not ret.type.startswith(".")
        ):
            # We need to know the bss space for the segment.
            if isinstance(parent, CommonSegCode):
                if parent.bss_size <= 0:
                    log.error(
                        f"Top-level segment '{parent.name}' is missing a `bss_size` value.\n    A non-zero `bss_size` value must be defined on the top-level segments that contain '{ret.type}' sections (produced by the '{ret.name}' section)."
                    )
                if (
                    isinstance(ret.vram_start, int)
                    and isinstance(parent.vram_end, int)
                    and ret.vram_start >= parent.vram_end
                ):
                    log.error(
                        f"The section '{ret.name}' (vram 0x{ret.vram_start:08X}) is outside its parent's address range '{parent.name}' (0x{parent.vram_start:08X} ~ 0x{parent.vram_end:08X}).\n    This may happen when the specified `bss_size` value is too small."
                    )

        ret.given_section_order = parse_segment_section_order(yaml)
        ret.given_subalign = parse_segment_subalign(yaml)

        if isinstance(yaml, dict):
            ret.extract = bool(yaml.get("extract", ret.extract))
            ret.exclusive_ram_id = yaml.get("exclusive_ram_id")
            ret.given_dir = Path(yaml.get("dir", ""))
            ret.has_linker_entry = bool(yaml.get("linker_entry", True))
            ret.given_find_file_boundaries = yaml.get("find_file_boundaries", None)

        ret.given_symbol_name_format = Segment.parse_segment_symbol_name_format(yaml)
        ret.given_symbol_name_format_no_rom = (
            Segment.parse_segment_symbol_name_format_no_rom(yaml)
        )
        ret.file_path = Segment.parse_segment_file_path(yaml)

        ret.bss_contains_common = Segment.parse_segment_bss_contains_common(
            yaml, options.opts.ld_bss_contains_common
        )

        ret.given_follows_vram = parse_segment_follows_vram(yaml)
        ret.given_vram_symbol = parse_segment_vram_symbol(yaml)

        if vram_class:
            ret.vram_class = vram_class
            if ret.given_follows_vram:
                log.error(
                    f"Error: segment {ret.name} has both a vram class and a follows_vram property"
                )
            if ret.given_vram_symbol:
                log.error(
                    f"Error: segment {ret.name} has both a vram class and a vram_symbol property"
                )

        if not ret.align:
            ret.align = parse_segment_align(yaml)
        return ret

    # For executable segments (.text); like c, asm or hasm
    @staticmethod
    def is_text() -> bool:
        return False

    # For read-write segments (.data); like data
    @staticmethod
    def is_data() -> bool:
        return False

    # For readonly segments (.rodata); like rodata or rdata
    @staticmethod
    def is_rodata() -> bool:
        return False

    # For segments which does not take space in ROM; like bss
    @staticmethod
    def is_noload() -> bool:
        return False

    @staticmethod
    def estimate_size(yaml: Union[Dict, List]) -> Optional[int]:
        return None

    @property
    def needs_symbols(self) -> bool:
        return False

    @property
    def dir(self) -> Path:
        if self.parent:
            return self.parent.dir / self.given_dir
        else:
            return self.given_dir

    @property
    def show_file_boundaries(self) -> bool:
        # If the segment has explicitly set `find_file_boundaries`, use it.
        if self.given_find_file_boundaries is not None:
            return self.given_find_file_boundaries

        # If the segment has no parent, use options as default.
        if not self.parent:
            return options.opts.find_file_boundaries

        return self.parent.show_file_boundaries

    @property
    def symbol_name_format(self) -> str:
        return self.given_symbol_name_format

    @property
    def symbol_name_format_no_rom(self) -> str:
        return self.given_symbol_name_format_no_rom

    @property
    def subalign(self) -> Optional[int]:
        assert self.parent is None, (
            f"subalign is not valid for non-top-level segments. ({self})"
        )
        return self.given_subalign

    @property
    def vram_symbol(self) -> Optional[str]:
        if self.vram_class and self.vram_class.vram_symbol:
            return self.vram_class.vram_symbol
        elif self.given_vram_symbol:
            return self.given_vram_symbol
        else:
            return None

    def get_exclusive_ram_id(self) -> Optional[str]:
        if self.parent:
            return self.parent.get_exclusive_ram_id()
        return self.exclusive_ram_id

    def add_symbol(self, symbol: Symbol):
        if symbol.vram_start not in self.given_seg_symbols:
            self.given_seg_symbols[symbol.vram_start] = []
        self.given_seg_symbols[symbol.vram_start].append(symbol)

        # For larger symbols, add their ranges to interval trees for faster lookup
        if symbol.size > 4:
            self.symbol_ranges_ram.addi(symbol.vram_start, symbol.vram_end, symbol)
            if symbol.rom is not None:
                self.symbol_ranges_rom.addi(symbol.rom, symbol.rom_end, symbol)

    @property
    def seg_symbols(self) -> Dict[int, List[Symbol]]:
        if self.parent:
            return self.parent.seg_symbols
        else:
            return self.given_seg_symbols

    @property
    def size(self) -> Optional[int]:
        if self.rom_start is not None and self.rom_end is not None:
            return self.rom_end - self.rom_start
        else:
            return None

    @property
    def statistics(self) -> SegmentStatistics:
        stats = empty_statistics()
        if self.size is not None:
            stats[self.statistics_type] = SegmentStatisticsInfo(size=self.size, count=1)
        return stats

    @property
    def statistics_type(self) -> SegmentType:
        return self.type

    @property
    def vram_end(self) -> Optional[int]:
        if self.vram_start is not None and self.size is not None:
            return self.vram_start + self.size
        else:
            return None

    @property
    def section_order(self) -> List[str]:
        return self.given_section_order

    @property
    def rodata_follows_data(self) -> bool:
        if ".rodata" not in self.section_order or ".data" not in self.section_order:
            return False
        return (
            self.section_order.index(".rodata") - self.section_order.index(".data") == 1
        )

    @property
    def suggestion_rodata_section_start(self) -> bool:
        if self.given_suggestion_rodata_section_start is not None:
            return self.given_suggestion_rodata_section_start
        if self.parent is not None:
            return self.parent.suggestion_rodata_section_start
        return options.opts.suggestion_rodata_section_start

    def get_cname(self) -> str:
        name = self.name
        if self.parent:
            name = self.parent.name + "_" + name

        return to_cname(name)

    def contains_vram(self, vram: int) -> bool:
        if self.vram_start is not None and self.vram_end is not None:
            return vram >= self.vram_start and vram < self.vram_end
        else:
            return False

    def contains_rom(self, rom: int) -> bool:
        if self.rom_start is not None and self.rom_end is not None:
            return rom >= self.rom_start and rom < self.rom_end
        else:
            return False

    def rom_to_ram(self, rom_addr: int) -> Optional[int]:
        if self.vram_start is not None and self.rom_start is not None:
            return self.vram_start + rom_addr - self.rom_start
        else:
            return None

    def ram_to_rom(self, ram_addr: int) -> Optional[int]:
        if not self.contains_vram(ram_addr) and ram_addr != self.vram_end:
            return None

        if self.vram_start is not None and self.rom_start is not None:
            return self.rom_start + ram_addr - self.vram_start
        else:
            return None

    def should_scan(self) -> bool:
        return self.should_split()

    def should_split(self) -> bool:
        return self.extract and options.opts.is_mode_active(self.type)

    def scan(self, rom_bytes: bytes):
        pass

    def split(self, rom_bytes: bytes):
        pass

    def cache(self):
        return (self.yaml, self.rom_end)

    def get_linker_section(self) -> str:
        return ".data"

    def get_linker_section_order(self) -> str:
        """
        Used to override the linking _order_ of a specific section

        Useful for files that may have non-conventional orderings (like putting .data with the other .rodata sections)
        """
        if self.linker_section_order is not None:
            return self.linker_section_order
        return self.get_linker_section()

    def get_linker_section_linksection(self) -> str:
        """
        The actual section that will be used when linking
        """
        if self.linker_section is not None:
            return self.linker_section
        return self.get_linker_section()

    def get_section_flags(self) -> Optional[str]:
        """
        Allows specifying flags for a section.

        This can be useful when creating a custom section, since sections not recognized by the linker will not be linked properly.

        GNU as docs about the section directive and flags: https://sourceware.org/binutils/docs/as/Section.html#ELF-Version

        Example:

        ```
        def get_section_flags(self) -> Optional[str]:
            # Tells the linker to allocate this section
            return "a"
        ```
        """
        return None

    def get_section_asm_line(self) -> str:
        line = f".section {self.get_linker_section_linksection()}"
        section_flags = self.get_section_flags()
        if section_flags:
            line += f', "{section_flags}"'
        return line

    def out_path(self) -> Optional[Path]:
        return None

    def get_most_parent(self) -> "Segment":
        seg = self

        while seg.parent:
            seg = seg.parent

        return seg

    def get_linker_entries(self) -> "List[LinkerEntry]":
        from ..segtypes.linker_entry import LinkerEntry

        if not self.has_linker_entry:
            return []

        path = self.out_path()

        if path:
            return [
                LinkerEntry(
                    self,
                    [path],
                    path,
                    self.get_linker_section_order(),
                    self.get_linker_section_linksection(),
                    self.is_noload(),
                )
            ]
        else:
            return []

    def log(self, msg):
        if options.opts.verbose:
            log.write(f"{self.type} {self.name}: {msg}")

    def warn(self, msg: str):
        self.warnings.append(msg)

    @staticmethod
    def get_default_name(addr) -> str:
        return f"{addr:X}"

    def is_name_default(self):
        return self.name == self.get_default_name(self.rom_start)

    def unique_id(self):
        if self.parent:
            s = self.parent.unique_id() + "_"
        else:
            s = ""

        return s + self.type + "_" + self.name

    @staticmethod
    def visible_ram(seg1: "Segment", seg2: "Segment") -> bool:
        if seg1.get_most_parent() == seg2.get_most_parent():
            return True
        if seg1.get_exclusive_ram_id() is None or seg2.get_exclusive_ram_id() is None:
            return True
        return seg1.get_exclusive_ram_id() != seg2.get_exclusive_ram_id()

    def retrieve_symbol(
        self, syms: Dict[int, List[Symbol]], addr: int
    ) -> Optional[Symbol]:
        if addr not in syms:
            return None

        items = syms[addr]

        # Filter out symbols that are in different top-level segments with the same unique_ram_id
        items = [
            i
            for i in items
            if i.segment is None or Segment.visible_ram(self, i.segment)
        ]

        if len(items) > 1:
            # print(f"Trying to retrieve {addr:X} from symbol dict but there are {len(items)} entries to pick from - picking the first")
            pass
        if len(items) == 0:
            return None
        return items[0]

    def retrieve_sym_type(
        self, syms: Dict[int, List[Symbol]], addr: int, type: str
    ) -> Optional[symbols.Symbol]:
        if addr not in syms:
            return None

        items = syms[addr]

        items = [
            i
            for i in items
            if (i.segment is None or Segment.visible_ram(self, i.segment))
            and (type == i.type)
        ]

        if len(items) == 0:
            return None

        return items[0]

    def get_symbol(
        self,
        addr: int,
        in_segment: bool = False,
        type: Optional[str] = None,
        create: bool = False,
        define: bool = False,
        reference: bool = False,
        search_ranges: bool = False,
        local_only: bool = False,
    ) -> Optional[Symbol]:
        ret: Optional[Symbol] = None
        rom: Optional[int] = None

        most_parent = self.get_most_parent()

        if in_segment:
            # If the vram address is within this segment, we can calculate the symbol's rom address
            rom = most_parent.ram_to_rom(addr)
            ret = most_parent.retrieve_symbol(most_parent.seg_symbols, addr)

            if not ret and search_ranges:
                # Search ranges first, starting with rom
                if rom is not None:
                    cands: Set[Interval] = most_parent.symbol_ranges_rom[rom]
                    if cands:
                        ret = cands.pop().data
                # and then vram if we can't find a rom match
                if not ret:
                    cands = most_parent.symbol_ranges_ram[addr]
                    if cands:
                        ret = cands.pop().data
        elif not local_only:
            ret = most_parent.retrieve_symbol(symbols.all_symbols_dict, addr)

            if not ret and search_ranges:
                cands = symbols.all_symbols_ranges[addr]
                if cands:
                    ret = cands.pop().data

        # Create the symbol if it doesn't exist
        if not ret and create:
            ret = Symbol(addr, rom=rom, type=type)
            symbols.add_symbol(ret)

            if in_segment:
                ret.segment = most_parent
                if addr not in most_parent.seg_symbols:
                    most_parent.seg_symbols[addr] = []
                most_parent.seg_symbols[addr].append(ret)

        if ret:
            if define:
                ret.defined = True
            if reference:
                ret.referenced = True
            if ret.type is None:
                ret.type = type
            if ret.rom is None:
                ret.rom = rom
            if in_segment:
                if ret.segment is None:
                    ret.segment = most_parent

        return ret

    def create_symbol(
        self,
        addr: int,
        in_segment: bool,
        type: Optional[str] = None,
        define: bool = False,
        reference: bool = False,
        search_ranges: bool = False,
        local_only: bool = False,
    ) -> Symbol:
        ret = self.get_symbol(
            addr,
            in_segment=in_segment,
            type=type,
            create=True,
            define=define,
            reference=reference,
            search_ranges=search_ranges,
            local_only=local_only,
        )
        assert ret is not None

        return ret

    def __repr__(self) -> str:
        # Shows a nicer string on the debugging screen
        return f"{self.name} ({self.type})"



from ..util import options, symbols


def init(target_bytes: bytes):
    symbols.spim_context.fillDefaultBannedSymbols()

    if options.opts.libultra_symbols:
        symbols.spim_context.globalSegment.fillLibultraSymbols()
    if options.opts.ique_symbols:
        symbols.spim_context.globalSegment.fillIQueSymbols()
    if options.opts.hardware_regs:
        symbols.spim_context.globalSegment.fillHardwareRegs(True)



from dataclasses import dataclass
import os
from pathlib import Path
from typing import cast, Dict, List, Literal, Mapping, Optional, Set, Type, TypeVar

from . import compiler
from .compiler import Compiler


@dataclass
class SplatOpts:
    # Debug / logging
    verbose: bool
    dump_symbols: bool
    modes: List[str]

    # Project configuration

    # Determines the base path of the project. Everything is relative to this path
    base_path: Path
    # Determines the path to the target binary
    target_path: Path
    # Path to the final elf target
    elf_path: Optional[Path]
    # Determines the platform of the target binary
    platform: str
    # Determines the compiler used to compile the target binary
    compiler: Compiler
    # Determines the endianness of the target binary
    endianness: Literal["big", "little"]
    # Determines the default section order of the target binary
    # this can be overridden per-segment
    section_order: List[str]
    # Determines the code that is inserted by default in generated .c files
    generated_c_preamble: str
    # Determines the code that is inserted by default in generated .s files
    generated_s_preamble: str
    # Determines any extra content to be added in the generated macro.inc file
    generated_macro_inc_content: Optional[str]
    # Determines if files related to assembly macros should be regenerated by splat
    generate_asm_macros_files: bool
    # Changes the definition of the generated `INCLUDE_ASM`.
    # default: The default one.
    # maspsx_hack: Use the maspsx hack workaround definition https://github.com/mkst/maspsx?tab=readme-ov-file#include_asm-reordering-workaround-hack
    include_asm_macro_style: Literal["default", "maspsx_hack"]
    # Directory to place the generated asm macros files.
    generated_asm_macros_directory: Path
    # Determines whether to use .o as the suffix for all binary files?... TODO document
    use_o_as_suffix: bool
    # the value of the $gp register to correctly calculate offset to %gp_rel relocs
    gp: Optional[int]
    # Checks and errors if there are any non consecutive segment types
    check_consecutive_segment_types: bool
    # Disable checks on `platform` option.
    is_unsupported_platform: bool
    # Allows to take precedence over the splat builtin platform segments via splat extension.
    allow_segment_overrides: bool

    # Paths
    asset_path: Path
    # Determines the path to the symbol addresses file(s)
    # A symbol_addrs file is to be updated/curated manually and contains addresses of symbols
    # as well as optional metadata such as rom address, type, and more
    #
    # It's possible to use more than one file by supplying a list instead of a string
    symbol_addrs_paths: List[Path]
    reloc_addrs_paths: List[Path]
    # Determines the path to the project build directory
    build_path: Path
    # Determines the path to the source code directory
    src_path: Path
    # Determines the path to the asm code directory
    asm_path: Path
    # Determines the path to the asm data directory
    data_path: Path
    # Determines the path to the asm nonmatchings directory
    nonmatchings_path: Path
    # Determines the path to the asm matchings directory (used alongside `disassemble_all` to organize matching functions from nonmatching functions)
    matchings_path: Path
    # Determines the path to the cache file (used when supplied --use-cache via the CLI)
    cache_path: Path
    # Tells splat to consider `hasm` files to be relative to `src_path` instead of `asm_path`.
    hasm_in_src_path: bool

    # Determines whether to create an automatically-generated undefined functions file
    # this file stores all functions that are referenced in the code but are not defined as seen by splat
    create_undefined_funcs_auto: bool
    # Determines the path to the undefined_funcs_auto file
    undefined_funcs_auto_path: Path

    # Determines whether to create an automatically-generated undefined symbols file
    # this file stores all symbols that are referenced in the code but are not defined as seen by splat
    create_undefined_syms_auto: bool
    # Determines the path to the undefined_symbols_auto file
    undefined_syms_auto_path: Path

    # Determines the path in which to search for custom splat extensions
    extensions_path: Optional[Path]

    # Determines the path to library files that are to be linked into the target binary
    lib_path: Path
    # Determines the path to object files that are to be linked into the target binary
    o_path: Path

    # TODO document
    elf_section_list_path: Optional[Path]

    # Linker script
    # Determines the default subalign value to be specified in the generated linker script
    subalign: Optional[int]
    # Determines whether to emit the subalign directive in the generated linker script
    emit_subalign: bool
    # The following option determines a list of sections for which automatic linker script entries should be added
    auto_link_sections: List[str]
    # Determines the desired path to the linker script that splat will generate
    ld_script_path: Path
    # Determines the desired path to the linker symbol header,
    # which exposes externed definitions for all segment ram/rom start/end locations
    ld_symbol_header_path: Optional[Path]
    # Determines whether to add a discard section with a wildcard to the linker script
    ld_discard_section: bool
    # A list of sections to preserve during link time. It can be useful to preserve debugging sections
    ld_sections_allowlist: List[str]
    # A list of sections to discard during link time. It can be useful to avoid using the wildcard discard. Note that this option does not turn off `ld_discard_section`
    ld_sections_denylist: List[str]
    # Determines whether to add wildcards for section linking in the linker script (.rodata* for example)
    ld_wildcard_sections: bool
    # Determines whether to use `follows_vram` (segment option) and
    # `vram_symbol` / `follows_classes` (vram_class options) to calculate vram addresses in the linker script.
    # If disabled, this uses the plain integer values for vram addresses defined in the yaml.
    ld_use_symbolic_vram_addresses: bool
    # Ensures segments of vram classes with dependencies (`vram_symbol` / `follows_classes`) are written to the linker script AFTER the segments they depend on.
    ld_sort_segments_by_vram_class_dependency: bool
    # Change linker script generation to allow partially linking segments. Requires both `ld_partial_scripts_path` and `ld_partial_build_segments_path` to be set.
    ld_partial_linking: bool
    # Folder were each intermediary linker script will be written to.
    ld_partial_scripts_path: Optional[Path]
    # Folder where the built partially linked segments will be placed by the build system.
    ld_partial_build_segments_path: Optional[Path]
    # Generate a dependency file for every linker script generated. Dependency files will have the same path and name as the corresponding linker script, but changing the extension to `.d`. Requires `elf_path` to be set.
    ld_dependencies: bool
    # Emit an `-include` directive for each object file in the dependency file.
    ld_dependencies_include: bool
    # Legacy linker script generation does not impose the section_order specified in the yaml options or per-segment options.
    ld_legacy_generation: bool
    # If enabled, the end symbol for each segment will be placed before the alignment directive for the segment
    segment_end_before_align: bool
    # Controls the style of the auto-generated segment symbols in the linker script. Possible values: splat, makerom
    segment_symbols_style: str
    # Specifies the starting offset for rom address symbols in the linker script.
    ld_rom_start: int
    # The value passed to the FILL statement on each segment. `None` disables using FILL statements on the linker script. Defaults to a fill value of 0.
    ld_fill_value: Optional[int]
    # Allows to control if `bss` sections (and derivatived sections) will be put on a `NOLOAD` segment on the generated linker script or not.
    ld_bss_is_noload: bool
    # Aligns the start of the segment to the given value
    ld_align_segment_start: Optional[int]
    # Allows to toggle aligning the `*_VRAM_END` linker symbol for each segment.
    ld_align_segment_vram_end: bool
    # Allows to toggle aligning the `*_END` linker symbol for each section of each section.
    ld_align_section_vram_end: bool
    # If enabled, the generated linker script will have a linker symbol for each data file
    ld_generate_symbol_per_data_segment: bool
    # Sets the default option for the `bss_contains_common` attribute of all segments.
    ld_bss_contains_common: bool
    # Specify an expression to be used for the `_gp` symbol in the generated linker script instead of a hardcoded value.
    ld_gp_expression: Optional[str]

    ################################################################################
    # C file options
    ################################################################################
    # Determines whether to create new c files if they don't exist
    create_c_files: bool
    # Determines whether to "auto-decompile" empty functions
    auto_decompile_empty_functions: bool
    # Determines whether to detect matched/unmatched functions in existing c files
    # so we can avoid creating .s files for already-decompiled functions
    do_c_func_detection: bool
    # Determines the newline char(s) to be used in c files
    c_newline: str

    ################################################################################
    # (Dis)assembly-related options
    ################################################################################
    # The following options determine the format that symbols should be named by default
    symbol_name_format: str
    # Same as above but for symbols with no rom address
    symbol_name_format_no_rom: str
    # Determines whether to detect and hint to the user about likely file splits when disassembling
    find_file_boundaries: bool
    # Determines whether to detect and hint to the user about possible rodata sections corresponding to a text section
    pair_rodata_to_text: bool
    # Determines whether to attempt to automatically migrate rodata into functions
    # (only works in certain circumstances)
    migrate_rodata_to_functions: bool
    # Determines the header to be used in every asm file that's included from c files
    asm_inc_header: str
    # Determines the macro used to declare functions in asm files
    asm_function_macro: str
    # Determines the macro used to declare symbols in the middle of functions in asm files (which may be alternative entries)
    asm_function_alt_macro: str
    # Determines the macro used to declare jumptable labels in asm files
    asm_jtbl_label_macro: str
    # Determines the macro used to declare data symbols in asm files
    asm_data_macro: str
    # Determines the macro used at the end of a function, such as endlabel or .end
    asm_end_label: str
    # Determines the macro used at the end of a data symbol, such as enddlabel
    asm_data_end_label: str
    # Determines the macro used to declare ehtable labels in asm files
    asm_ehtable_label_macro: str
    # Determines the macro used to declare the given symbol is a non matching one.
    asm_nonmatching_label_macro: str
    # Toggles the .size directive emitted by the disassembler
    asm_emit_size_directive: Optional[bool]
    # Determines the number of characters to left align before the TODO finish documenting
    mnemonic_ljust: int
    # Determines whether to pad the rom address
    rom_address_padding: bool
    # Determines which ABI names to use for general purpose registers
    mips_abi_gpr: str
    # Determines which ABI names to use for floating point registers
    # Valid values: 'numeric', 'o32', 'n32', 'n64'
    # o32 is highly recommended, as it provides logically named registers for floating point instructions
    # For more info, see https://gist.github.com/EllipticEllipsis/27eef11205c7a59d8ea85632bc49224d
    mips_abi_float_regs: str
    # Determines whether functions inside c files should have named registers
    named_regs_for_c_funcs: bool
    # Determines whether to add ".set gp=64" to asm/hasm files
    add_set_gp_64: bool
    # Generate .asmproc.d dependency files for each C file which still reference functions in assembly files
    create_asm_dependencies: bool
    # Global option for rodata string encoding. This can be overriden per segment
    string_encoding: Optional[str]
    # Global option for data string encoding. This can be overriden per segment
    data_string_encoding: Optional[str]
    # Global option for the rodata string guesser. 0 disables the guesser completely.
    rodata_string_guesser_level: Optional[int]
    # Global option for the data string guesser. 0 disables the guesser completely.
    data_string_guesser_level: Optional[int]
    # Global option for allowing data symbols using addends on symbol references. It can be overriden per symbol
    allow_data_addends: bool
    # Tells the disassembler to try disassembling functions with unknown instructions instead of falling back to disassembling as raw data
    disasm_unknown: bool
    # Tries to detect redundant and unreferenced functions ends and merge them together. This option is ignored if the compiler is not set to IDO.
    detect_redundant_function_end: bool
    # Don't skip disassembling already matched functions and migrated sections
    disassemble_all: bool
    # Emit a full `.s` file for each `c`/`cpp` segment besides the generated `nonmatchings` individual functions
    make_full_disasm_for_code: bool
    # Allow specifying that the global memory range may be larger than what was automatically detected.
    # Useful for projects where splat is used in multiple individual files, meaning the expected global segment may not be properly detected because each instance of splat can't see the info from other files.
    global_vram_start: Optional[int]
    global_vram_end: Optional[int]
    # For `c` segments (functions under the nonmatchings folder).
    # If True then use the `%gp_rel` explicit relocation parameter on instructions that use the $gp register,
    # otherwise strip the `%gp_rel` parameter entirely and convert those instructions into macro instructions that may not assemble to the original
    # bytes. In the latter case, it is the user's responsability to provide the symbol's information to the assembler so it can assemble the
    # instruction with the proper relocation.
    use_gp_rel_macro_nonmatching: bool
    # Does the same as `use_gp_rel_macro_nonmatching`, except it is only applied to `asm` and `hasm` segments.
    use_gp_rel_macro: bool
    # Allows emitting suggestions for where the rodata may start by examining the data section.
    suggestion_rodata_section_start: bool

    ################################################################################
    # N64-specific options
    ################################################################################
    # Determines the encoding of the header
    header_encoding: str
    # Determines the type gfx ucode (used by gfx segments)
    # Valid options are ['f3d', 'f3db', 'f3dex', 'f3dexb', 'f3dex2']
    gfx_ucode: str
    # Use named libultra symbols by default. Those will need to be added to a linker script manually by the user
    libultra_symbols: bool
    # Use named libultra symbols by default. Those will need to be added to a linker script manually by the user
    ique_symbols: bool
    # Use named hardware register symbols by default. Those will need to be added to a linker script manually by the user
    hardware_regs: bool
    # Append the image type to the output file extension
    image_type_in_extension: bool

    ################################################################################
    # Compiler-specific options
    ################################################################################
    # Determines whether to use a legacy INCLUDE_ASM macro format in c files
    # only applies to GCC/SN64
    use_legacy_include_asm: bool
    # Emit alignment directives in branch labels, as a way to workaround the short loop bug present in SN PS2 compilers
    align_on_branch_labels: bool

    # Returns whether the given mode is currently enabled
    def is_mode_active(self, mode: str) -> bool:
        return mode in self.modes or "all" in self.modes


opts: SplatOpts


T = TypeVar("T")


class OptParser:
    _read_opts: Set[str]

    def __init__(self, yaml: Mapping[str, object]) -> None:
        self._yaml = yaml
        self._read_opts = set()

    def parse_opt(self, opt: str, t: Type[T], default: Optional[T] = None) -> T:
        if opt not in self._yaml:
            if default is not None:
                return default
            raise ValueError(f"Missing required option {opt}")
        self._read_opts.add(opt)
        value = self._yaml[opt]
        if isinstance(value, t):
            return value
        if t is float and isinstance(value, int):
            return cast(T, float(value))
        raise ValueError(f"Expected {opt} to have type {t}, got {type(value)}")

    def parse_optional_opt(self, opt: str, t: Type[T]) -> Optional[T]:
        if opt not in self._yaml:
            return None
        return self.parse_opt(opt, t)

    def parse_optional_opt_with_default(
        self, opt: str, t: Type[T], default: Optional[T]
    ) -> Optional[T]:
        if opt not in self._yaml:
            return default
        self._read_opts.add(opt)
        value = self._yaml[opt]
        if value is None or isinstance(value, t):
            return value
        if t is float and isinstance(value, int):
            return cast(T, float(value))
        raise ValueError(f"Expected {opt} to have type {t}, got {type(value)}")

    def parse_opt_within(
        self, opt: str, t: Type[T], within: List[T], default: Optional[T] = None
    ) -> T:
        value = self.parse_opt(opt, t, default)
        if value not in within:
            raise ValueError(f"Invalid value for {opt}: {value}")
        return value

    def parse_path(
        self, base_path: Path, opt: str, default: Optional[str] = None
    ) -> Path:
        return Path(os.path.normpath(base_path / self.parse_opt(opt, str, default)))

    def parse_optional_path(self, base_path: Path, opt: str) -> Optional[Path]:
        if opt not in self._yaml:
            return None
        return self.parse_path(base_path, opt)

    def parse_path_list(self, base_path: Path, opt: str, default: str) -> List[Path]:
        paths = self.parse_opt(opt, object, default)

        if isinstance(paths, str):
            return [base_path / paths]
        elif isinstance(paths, list):
            return [base_path / path for path in paths]
        else:
            raise ValueError(f"Expected str or list for '{opt}', got {type(paths)}")

    def check_no_unread_opts(self) -> None:
        opts = [opt for opt in self._yaml if opt not in self._read_opts]
        if opts:
            raise ValueError(f"Unrecognized YAML option(s): {', '.join(opts)}")


def _parse_yaml(
    yaml: Dict,
    config_paths: List[Path],
    modes: List[str],
    verbose: bool = False,
    disasm_all: bool = False,
    make_full_disasm_for_code: bool = False,
) -> SplatOpts:
    p = OptParser(yaml)

    basename = p.parse_opt("basename", str)
    is_unsupported_platform = p.parse_opt("is_unsupported_platform", bool, False)

    if is_unsupported_platform:
        platform = p.parse_opt("platform", str)
    else:
        platform = p.parse_opt_within("platform", str, ["n64", "psx", "ps2", "psp"])

    comp = compiler.for_name(p.parse_opt("compiler", str, "IDO"))

    base_path = Path(
        os.path.normpath(config_paths[0].parent / p.parse_opt("base_path", str))
    )
    asm_path: Path = p.parse_path(base_path, "asm_path", "asm")

    asm_emit_size_directive = p.parse_optional_opt("asm_emit_size_directive", bool)
    # If option not provided then use the compiler default
    if asm_emit_size_directive is None:
        asm_emit_size_directive = comp.asm_emit_size_directive

    align_on_branch_labels = p.parse_optional_opt("align_on_branch_labels", bool)
    # If option not provided then use the compiler default
    if align_on_branch_labels is None:
        align_on_branch_labels = comp.align_on_branch_labels

    def parse_endianness() -> Literal["big", "little"]:
        endianness = p.parse_opt_within(
            "endianness",
            str,
            ["big", "little"],
            "little" if platform in ["psx", "ps2", "psp"] else "big",
        )

        if endianness == "big":
            return "big"
        elif endianness == "little":
            return "little"
        else:
            raise ValueError(f"Invalid endianness: {endianness}")

    def parse_include_asm_macro_style() -> Literal["default", "maspsx_hack"]:
        include_asm_macro_style = p.parse_opt_within(
            "include_asm_macro_style",
            str,
            ["default", "maspsx_hack"],
            "default",
        )

        if include_asm_macro_style == "default":
            return "default"
        elif include_asm_macro_style == "maspsx_hack":
            return "maspsx_hack"
        else:
            raise ValueError(f"Invalid endianness: {include_asm_macro_style}")

    default_ld_bss_is_noload = True
    if platform == "psx":
        default_ld_bss_is_noload = False

    default_add_set_gp_64 = True
    if platform in ("psx", "ps2"):
        default_add_set_gp_64 = False

    ret = SplatOpts(
        verbose=verbose,
        dump_symbols=p.parse_opt("dump_symbols", bool, False),
        modes=modes,
        base_path=base_path,
        target_path=p.parse_path(base_path, "target_path"),
        elf_path=p.parse_optional_path(base_path, "elf_path"),
        platform=platform,
        is_unsupported_platform=is_unsupported_platform,
        allow_segment_overrides=p.parse_opt("allow_segment_overrides", bool, False),
        compiler=comp,
        endianness=parse_endianness(),
        section_order=p.parse_opt(
            "section_order", list, [".text", ".data", ".rodata", ".bss"]
        ),
        generated_c_preamble=p.parse_opt(
            "generated_c_preamble", str, '#include "common.h"'
        ),
        generated_s_preamble=p.parse_opt("generated_s_preamble", str, ""),
        generated_macro_inc_content=p.parse_optional_opt(
            "generated_macro_inc_content", str
        ),
        generate_asm_macros_files=p.parse_opt("generate_asm_macros_files", bool, True),
        include_asm_macro_style=parse_include_asm_macro_style(),
        generated_asm_macros_directory=p.parse_path(
            base_path, "generated_asm_macros_directory", "include"
        ),
        use_o_as_suffix=p.parse_opt("o_as_suffix", bool, False),
        gp=p.parse_optional_opt("gp_value", int),
        check_consecutive_segment_types=p.parse_opt(
            "check_consecutive_segment_types", bool, True
        ),
        asset_path=p.parse_path(base_path, "asset_path", "assets"),
        symbol_addrs_paths=p.parse_path_list(
            base_path, "symbol_addrs_path", "symbol_addrs.txt"
        ),
        reloc_addrs_paths=p.parse_path_list(
            base_path, "reloc_addrs_path", "reloc_addrs.txt"
        ),
        build_path=p.parse_path(base_path, "build_path", "build"),
        src_path=p.parse_path(base_path, "src_path", "src"),
        asm_path=asm_path,
        data_path=p.parse_path(asm_path, "data_path", "data"),
        nonmatchings_path=p.parse_path(asm_path, "nonmatchings_path", "nonmatchings"),
        matchings_path=p.parse_path(asm_path, "matchings_path", "matchings"),
        cache_path=p.parse_path(base_path, "cache_path", ".splache"),
        hasm_in_src_path=p.parse_opt("hasm_in_src_path", bool, False),
        create_undefined_funcs_auto=p.parse_opt(
            "create_undefined_funcs_auto", bool, True
        ),
        undefined_funcs_auto_path=p.parse_path(
            base_path, "undefined_funcs_auto_path", "undefined_funcs_auto.txt"
        ),
        create_undefined_syms_auto=p.parse_opt(
            "create_undefined_syms_auto", bool, True
        ),
        undefined_syms_auto_path=p.parse_path(
            base_path, "undefined_syms_auto_path", "undefined_syms_auto.txt"
        ),
        extensions_path=p.parse_optional_path(base_path, "extensions_path"),
        lib_path=p.parse_path(base_path, "lib_path", "lib"),
        o_path=p.parse_path(base_path, "o_path", "build"),
        elf_section_list_path=p.parse_optional_path(base_path, "elf_section_list_path"),
        subalign=p.parse_optional_opt_with_default("subalign", int, 16),
        emit_subalign=p.parse_opt("emit_subalign", bool, True),
        auto_link_sections=p.parse_opt(
            "auto_link_sections", list, [".data", ".rodata", ".bss"]
        ),
        ld_script_path=p.parse_path(base_path, "ld_script_path", f"{basename}.ld"),
        ld_symbol_header_path=p.parse_optional_path(base_path, "ld_symbol_header_path"),
        ld_discard_section=p.parse_opt("ld_discard_section", bool, True),
        ld_sections_allowlist=p.parse_opt("ld_sections_allowlist", list, []),
        ld_sections_denylist=p.parse_opt("ld_sections_denylist", list, []),
        ld_wildcard_sections=p.parse_opt("ld_wildcard_sections", bool, False),
        ld_sort_segments_by_vram_class_dependency=p.parse_opt(
            "ld_sort_segments_by_vram_class_dependency", bool, False
        ),
        ld_use_symbolic_vram_addresses=p.parse_opt(
            "ld_use_symbolic_vram_addresses", bool, True
        ),
        ld_partial_linking=p.parse_opt("ld_partial_linking", bool, False),
        ld_partial_scripts_path=p.parse_optional_path(
            base_path, "ld_partial_scripts_path"
        ),
        ld_partial_build_segments_path=p.parse_optional_path(
            base_path, "ld_partial_build_segments_path"
        ),
        ld_dependencies=p.parse_opt("ld_dependencies", bool, False),
        ld_dependencies_include=p.parse_opt("ld_dependencies_include", bool, True),
        ld_legacy_generation=p.parse_opt("ld_legacy_generation", bool, False),
        segment_end_before_align=p.parse_opt("segment_end_before_align", bool, False),
        segment_symbols_style=p.parse_opt_within(
            "segment_symbols_style", str, ["splat", "makerom"], "splat"
        ),
        ld_rom_start=p.parse_opt("ld_rom_start", int, 0),
        ld_fill_value=p.parse_optional_opt_with_default("ld_fill_value", int, 0),
        ld_bss_is_noload=p.parse_opt(
            "ld_bss_is_noload", bool, default_ld_bss_is_noload
        ),
        ld_align_segment_start=p.parse_optional_opt_with_default(
            "ld_align_segment_start", int, None
        ),
        ld_align_segment_vram_end=p.parse_opt("ld_align_segment_vram_end", bool, True),
        ld_align_section_vram_end=p.parse_opt("ld_align_section_vram_end", bool, True),
        ld_generate_symbol_per_data_segment=p.parse_opt(
            "ld_generate_symbol_per_data_segment", bool, False
        ),
        ld_bss_contains_common=p.parse_opt("ld_bss_contains_common", bool, False),
        ld_gp_expression=p.parse_optional_opt_with_default(
            "ld_gp_expression", str, None
        ),
        create_c_files=p.parse_opt("create_c_files", bool, True),
        auto_decompile_empty_functions=p.parse_opt(
            "auto_decompile_empty_functions", bool, True
        ),
        do_c_func_detection=p.parse_opt("do_c_func_detection", bool, True),
        c_newline=p.parse_opt("c_newline", str, comp.c_newline),
        symbol_name_format=p.parse_opt("symbol_name_format", str, "$VRAM"),
        symbol_name_format_no_rom=p.parse_opt(
            "symbol_name_format_no_rom", str, "$VRAM_$SEG"
        ),
        find_file_boundaries=p.parse_opt("find_file_boundaries", bool, True),
        pair_rodata_to_text=p.parse_opt("pair_rodata_to_text", bool, True),
        migrate_rodata_to_functions=p.parse_opt(
            "migrate_rodata_to_functions", bool, True
        ),
        asm_inc_header=p.parse_opt("asm_inc_header", str, comp.asm_inc_header),
        asm_function_macro=p.parse_opt(
            "asm_function_macro", str, comp.asm_function_macro
        ),
        asm_function_alt_macro=p.parse_opt(
            "asm_function_alt_macro", str, comp.asm_function_alt_macro
        ),
        asm_jtbl_label_macro=p.parse_opt(
            "asm_jtbl_label_macro", str, comp.asm_jtbl_label_macro
        ),
        asm_data_macro=p.parse_opt("asm_data_macro", str, comp.asm_data_macro),
        asm_end_label=p.parse_opt("asm_end_label", str, comp.asm_end_label),
        asm_data_end_label=p.parse_opt(
            "asm_data_end_label", str, comp.asm_data_end_label
        ),
        asm_ehtable_label_macro=p.parse_opt(
            "asm_ehtable_label_macro", str, comp.asm_ehtable_label_macro
        ),
        asm_nonmatching_label_macro=p.parse_opt(
            "asm_nonmatching_label_macro", str, comp.asm_nonmatching_label_macro
        ),
        asm_emit_size_directive=asm_emit_size_directive,
        mnemonic_ljust=p.parse_opt("mnemonic_ljust", int, 11),
        rom_address_padding=p.parse_opt("rom_address_padding", bool, False),
        mips_abi_gpr=p.parse_opt_within(
            "mips_abi_gpr",
            str,
            ["numeric", "32", "o32", "n32", "n64"],
            "o32",
        ),
        mips_abi_float_regs=p.parse_opt_within(
            "mips_abi_float_regs",
            str,
            ["numeric", "32", "o32", "n32", "n64"],
            "numeric",
        ),
        named_regs_for_c_funcs=p.parse_opt("named_regs_for_c_funcs", bool, True),
        add_set_gp_64=p.parse_opt("add_set_gp_64", bool, default_add_set_gp_64),
        create_asm_dependencies=p.parse_opt("create_asm_dependencies", bool, False),
        string_encoding=p.parse_optional_opt("string_encoding", str),
        data_string_encoding=p.parse_optional_opt("data_string_encoding", str),
        rodata_string_guesser_level=p.parse_optional_opt(
            "rodata_string_guesser_level", int
        ),
        data_string_guesser_level=p.parse_optional_opt(
            "data_string_guesser_level", int
        ),
        allow_data_addends=p.parse_opt("allow_data_addends", bool, True),
        header_encoding=p.parse_opt("header_encoding", str, "ASCII"),
        gfx_ucode=p.parse_opt_within(
            "gfx_ucode",
            str,
            ["f3d", "f3db", "f3dex", "f3dexb", "f3dex2"],
            "f3dex2",
        ),
        libultra_symbols=p.parse_opt("libultra_symbols", bool, False),
        ique_symbols=p.parse_opt("ique_symbols", bool, False),
        hardware_regs=p.parse_opt("hardware_regs", bool, False),
        image_type_in_extension=p.parse_opt("image_type_in_extension", bool, False),
        use_legacy_include_asm=p.parse_opt("use_legacy_include_asm", bool, False),
        align_on_branch_labels=align_on_branch_labels,
        disasm_unknown=p.parse_opt("disasm_unknown", bool, False),
        detect_redundant_function_end=p.parse_opt(
            "detect_redundant_function_end", bool, True
        ),
        # Setting either option will produce a full disassembly,
        # but we still have to check the yaml option first to avoid leaving option unparsed,
        # because splat would complain about an unrecognized yaml option otherwise.
        disassemble_all=p.parse_opt("disassemble_all", bool, False) or disasm_all,
        make_full_disasm_for_code=p.parse_opt("make_full_disasm_for_code", bool, False)
        or make_full_disasm_for_code,
        global_vram_start=p.parse_optional_opt("global_vram_start", int),
        global_vram_end=p.parse_optional_opt("global_vram_end", int),
        use_gp_rel_macro_nonmatching=p.parse_opt(
            "use_gp_rel_macro_nonmatching", bool, True
        ),
        use_gp_rel_macro=p.parse_opt("use_gp_rel_macro", bool, True),
        suggestion_rodata_section_start=p.parse_opt(
            "suggestion_rodata_section_start", bool, True
        ),
    )
    p.check_no_unread_opts()
    return ret


def initialize(
    config: Dict,
    config_paths: List[Path],
    modes: Optional[List[str]] = None,
    verbose=False,
    disasm_all=False,
    make_full_disasm_for_code=False,
):
    global opts

    if not modes:
        modes = ["all"]

    opts = _parse_yaml(
        config["options"],
        config_paths,
        modes,
        verbose,
        disasm_all,
        make_full_disasm_for_code,
    )



from . import cache_handler as cache_handler
from . import color as color
from . import compiler as compiler
from . import file_presets as file_presets
from . import log as log
from . import n64 as n64
from . import options as options
from . import palettes as palettes
from . import progress_bar as progress_bar
from . import ps2 as ps2
from . import psx as psx
from . import relocs as relocs
from . import statistics as statistics
from . import symbols as symbols
from . import utils as utils
from . import vram_classes as vram_classes



import spimdisasm
import rabbitizer


def init(target_bytes: bytes):
    rabbitizer.config.toolchainTweaks_treatJAsUnconditionalBranch = False

    spimdisasm.common.GlobalConfig.ABI = spimdisasm.common.Abi.EABI64
    spimdisasm.common.GlobalConfig.SYMBOL_ALIGNMENT_REQUIRES_ALIGNED_SECTION = True



import n64img.image

from .img import N64SegImg


class N64SegI1(N64SegImg):
    def __init__(self, *args, **kwargs):
        kwargs["img_cls"] = n64img.image.I1
        super().__init__(*args, **kwargs)



from dataclasses import dataclass
from typing import Optional, Dict


@dataclass
class Compiler:
    name: str
    asm_function_macro: str = "glabel"
    asm_function_alt_macro: str = "alabel"
    asm_jtbl_label_macro: str = "jlabel"
    asm_data_macro: str = "dlabel"
    asm_end_label: str = "endlabel"
    asm_data_end_label: str = "enddlabel"
    asm_ehtable_label_macro: str = "ehlabel"
    asm_nonmatching_label_macro: str = "nonmatching"
    c_newline: str = "\n"
    asm_inc_header: str = ""
    asm_emit_size_directive: Optional[bool] = None
    j_as_branch: bool = False
    uses_include_asm: bool = True
    align_on_branch_labels: bool = False


GCC = Compiler(
    "GCC",
    asm_inc_header=".set noat      /* allow manual use of $at */\n.set noreorder /* don't insert nops after branches */\n\n",
    j_as_branch=True,
)

SN64 = Compiler(
    "SN64",
    asm_function_macro=".globl",
    asm_function_alt_macro=".globl",
    asm_jtbl_label_macro=".globl",
    asm_data_macro=".globl",
    asm_end_label=".end",
    asm_data_end_label="",
    asm_nonmatching_label_macro="",
    c_newline="\r\n",
    asm_emit_size_directive=False,
    j_as_branch=True,
)

IDO = Compiler("IDO", asm_emit_size_directive=False, uses_include_asm=False)

KMC = Compiler(
    "KMC",
    j_as_branch=True,
)

# iQue
EGCS = Compiler(
    "EGCS",
    j_as_branch=True,
)

# PS1
PSYQ = Compiler(
    "PSYQ",
    j_as_branch=True,
)

# PS2
MWCCPS2 = Compiler("MWCCPS2", uses_include_asm=False)
EEGCC = Compiler("EEGCC", align_on_branch_labels=True)

compiler_for_name: Dict[str, Compiler] = {
    x.name: x
    for x in [
        GCC,
        SN64,
        IDO,
        KMC,
        EGCS,
        PSYQ,
        MWCCPS2,
        EEGCC,
    ]
}


def for_name(name: str) -> Compiler:
    name = name.upper()
    return compiler_for_name.get(name, Compiler(name))



from pathlib import Path
from typing import Optional, List
from ...util import options, symbols, log

from .codesubsegment import CommonSegCodeSubsegment
from .group import CommonSegGroup

from ...disassembler.disassembler_section import DisassemblerSection, make_data_section


class CommonSegData(CommonSegCodeSubsegment, CommonSegGroup):
    @staticmethod
    def is_data() -> bool:
        return True

    def asm_out_path(self) -> Path:
        typ = self.type
        if typ.startswith("."):
            typ = typ[1:]

        return options.opts.data_path / self.dir / f"{self.name}.{typ}.s"

    def out_path(self) -> Optional[Path]:
        if self.type.startswith("."):
            if self.sibling:
                # C file
                return self.sibling.out_path()
            else:
                # Implied C file
                return options.opts.src_path / self.dir / f"{self.name}.c"
        else:
            # ASM
            return self.asm_out_path()

    def scan(self, rom_bytes: bytes):
        CommonSegGroup.scan(self, rom_bytes)

        if self.rom_start is not None and self.rom_end is not None:
            self.disassemble_data(rom_bytes)

    def get_asm_file_extra_directives(self) -> List[str]:
        return []

    def split(self, rom_bytes: bytes):
        super().split(rom_bytes)

        if self.spim_section is None or not self.should_self_split():
            return

        self.split_as_asm_file(self.asm_out_path())

    def should_self_split(self) -> bool:
        return options.opts.is_mode_active("data") and (
            not self.type.startswith(".") or options.opts.disassemble_all
        )

    def should_scan(self) -> bool:
        return True

    def should_split(self) -> bool:
        return True

    def cache(self):
        return [CommonSegCodeSubsegment.cache(self), CommonSegGroup.cache(self)]

    def get_linker_section(self) -> str:
        return ".data"

    def get_section_flags(self) -> Optional[str]:
        return "wa"

    def get_linker_entries(self):
        return CommonSegCodeSubsegment.get_linker_entries(self)

    def configure_disassembler_section(
        self, disassembler_section: DisassemblerSection
    ) -> None:
        "Allows to configure the section before running the analysis on it"

        section = disassembler_section.get_section()

        # Set data string encoding
        # First check the global configuration
        if options.opts.data_string_encoding is not None:
            section.stringEncoding = options.opts.data_string_encoding

        # Then check the per-segment configuration in case we want to override the global one
        if self.str_encoding is not None:
            section.stringEncoding = self.str_encoding

    def disassemble_data(self, rom_bytes):
        if self.is_auto_segment:
            return

        if not isinstance(self.rom_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a rom_start. Got '{self.rom_start}'"
            )

        # Supposedly logic error, not user error
        assert isinstance(self.rom_end, int), self.rom_end

        # Supposedly logic error, not user error
        segment_rom_start = self.get_most_parent().rom_start
        assert isinstance(segment_rom_start, int), segment_rom_start

        if not isinstance(self.vram_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a vram address. Got '{self.vram_start}'"
            )

        self.spim_section = make_data_section(
            self.rom_start,
            self.rom_end,
            self.vram_start,
            self.name,
            rom_bytes,
            segment_rom_start,
            self.get_exclusive_ram_id(),
        )

        assert self.spim_section is not None

        self.configure_disassembler_section(self.spim_section)

        self.spim_section.analyze()
        self.spim_section.set_comment_offset(self.rom_start)

        rodata_encountered = False

        for symbol in self.spim_section.get_section().symbolList:
            symbols.create_symbol_from_spim_symbol(
                self.get_most_parent(), symbol.contextSym, force_in_segment=True
            )

            # Gather symbols found by spimdisasm and create those symbols in splat's side
            for referenced_vram in symbol.referencedVrams:
                context_sym = self.spim_section.get_section().getSymbol(
                    referenced_vram, tryPlusOffset=False
                )
                if context_sym is not None:
                    symbols.create_symbol_from_spim_symbol(
                        self.get_most_parent(), context_sym, force_in_segment=False
                    )

            # Hint to the user that we are now in the .rodata section and no longer in the .data section (assuming rodata follows data)
            if (
                self.suggestion_rodata_section_start
                and not rodata_encountered
                and self.get_most_parent().rodata_follows_data
            ):
                if symbol.contextSym.isJumpTable():
                    rodata_encountered = True
                    print(
                        f"\n\nData segment {self.name}, symbol at vram {symbol.contextSym.vram:X} is a jumptable, indicating the start of the rodata section _may_ be near here."
                    )
                    print(
                        "Please note the real start of the rodata section may be way before this point."
                    )
                    if symbol.contextSym.vromAddress is not None:
                        print(f"      - [0x{symbol.contextSym.vromAddress:X}, rodata]")



"""
Utilities to write assembly-related macros.
This includes writing files like:
- include/include_asm.h
- include/macro.inc
- include/labels.inc
- include/gte_macros.inc

The directory where these files are written to can be controlled with
`options.opts.generated_asm_macros_directory`.
"""

from pathlib import Path

from . import options, log


def write_all_files():
    if not options.opts.generate_asm_macros_files:
        return

    write_include_asm_h()
    write_assembly_inc_files()


def _write(filepath: str, contents: str):
    p = Path(filepath)
    p.parent.mkdir(parents=True, exist_ok=True)

    if p.exists():
        existing_contents = p.read_text(encoding="utf-8")
        if existing_contents == contents:
            return
    with p.open("w", encoding="utf-8", newline="\n") as f:
        f.write(contents)


def write_include_asm_h():
    if not options.opts.compiler.uses_include_asm:
        # These compilers do not use the `INCLUDE_ASM` macro.
        return

    directory = options.opts.generated_asm_macros_directory.as_posix()

    if options.opts.include_asm_macro_style == "maspsx_hack":
        include_asm_macro = """\
#define INCLUDE_ASM(FOLDER, NAME) \\
    void __maspsx_include_asm_hack_##NAME() { \\
        __asm__( \\
            ".text # maspsx-keep \\n" \\
            "    .align 2 # maspsx-keep\\n" \\
            "    .set noat # maspsx-keep\\n" \\
            "    .set noreorder # maspsx-keep\\n" \\
            "    .include \\"" FOLDER "/" #NAME ".s\\" # maspsx-keep\\n" \\
            "    .set reorder # maspsx-keep\\n" \\
            "    .set at # maspsx-keep\\n" \\
        ); \\
    }
"""
    else:  # default
        include_asm_macro = """\
#define INCLUDE_ASM(FOLDER, NAME) \\
    __asm__( \\
        ".section .text\\n" \\
        "    .set noat\\n" \\
        "    .set noreorder\\n" \\
        "    .include \\"" FOLDER "/" #NAME ".s\\"\\n" \\
        "    .set reorder\\n" \\
        "    .set at\\n" \\
    )
"""

    include_rodata_macro = """\
#define INCLUDE_RODATA(FOLDER, NAME) \\
    __asm__( \\
        ".section .rodata\\n" \\
        "    .include \\"" FOLDER "/" #NAME ".s\\"\\n" \\
        ".section .text" \\
    )
"""

    file_data = f"""\
#ifndef INCLUDE_ASM_H
#define INCLUDE_ASM_H

#if !defined(M2CTX) && !defined(PERMUTER)

#ifndef INCLUDE_ASM
{include_asm_macro}\
#endif
#ifndef INCLUDE_RODATA
{include_rodata_macro}\
#endif

#if INCLUDE_ASM_USE_MACRO_INC
__asm__(".include \\"{directory}/macro.inc\\"\\n");
#else
__asm__(".include \\"{directory}/labels.inc\\"\\n");
#endif

#else

#ifndef INCLUDE_ASM
#define INCLUDE_ASM(FOLDER, NAME)
#endif
#ifndef INCLUDE_RODATA
#define INCLUDE_RODATA(FOLDER, NAME)
#endif

#endif /* !defined(M2CTX) && !defined(PERMUTER) */

#endif /* INCLUDE_ASM_H */
"""
    _write(f"{directory}/include_asm.h", file_data)


def write_assembly_inc_files():
    directory = options.opts.generated_asm_macros_directory.as_posix()

    func_macros = f"""\
# A function symbol.
.macro {options.opts.asm_function_macro} label, visibility=global
    .\\visibility \\label
    .type \\label, @function
    \\label:
        .ent \\label
.endm
"""
    if options.opts.asm_end_label != "":
        func_macros += f"""
# The end of a function symbol.
.macro {options.opts.asm_end_label} label
    .size \\label, . - \\label
    .end \\label
.endm
"""
    if (
        options.opts.asm_function_alt_macro != ""
        and options.opts.asm_function_alt_macro != options.opts.asm_function_macro
    ):
        func_macros += f"""
# An alternative entry to a function.
.macro {options.opts.asm_function_alt_macro} label, visibility=global
    .\\visibility \\label
    .type \\label, @function
    \\label:
        .aent \\label
.endm
"""
    if (
        options.opts.asm_ehtable_label_macro != ""
        and options.opts.asm_ehtable_label_macro != options.opts.asm_function_macro
    ):
        func_macros += f"""
# A label referenced by an error handler table.
.macro {options.opts.asm_ehtable_label_macro} label, visibility=global
    .\\visibility \\label
    \\label:
.endm
"""

    jlabel_macro_labelsinc = ""
    jlabel_macro_macroinc = ""
    if (
        options.opts.asm_jtbl_label_macro != ""
        and options.opts.asm_jtbl_label_macro != options.opts.asm_function_macro
    ):
        jlabel_macro_macroinc = f"""
# A label referenced by a jumptable.
.macro {options.opts.asm_jtbl_label_macro} label, visibility=global
    .\\visibility \\label
    \\label:
.endm
"""
        if options.opts.migrate_rodata_to_functions:
            jlabel_macro_labelsinc = f"""
# A label referenced by a jumptable.
.macro {options.opts.asm_jtbl_label_macro} label, visibility=local
    .\\visibility \\label
    \\label:
.endm
"""
        else:
            # If the user doesn't migrate rodata, like jumptables, to functions
            # then the user will need jlabels to be global instead of local,
            # so we just reuse the definition from macro.inc
            jlabel_macro_labelsinc = jlabel_macro_macroinc

    data_macros = ""
    if (
        options.opts.asm_data_macro != ""
        and options.opts.asm_data_macro != options.opts.asm_function_macro
    ):
        data_macros += f"""
# A data symbol.
.macro {options.opts.asm_data_macro} label, visibility=global
    .\\visibility \\label
    .type \\label, @object
    \\label:
.endm
"""
    if options.opts.asm_data_end_label != "":
        data_macros += f"""
# End of a data symbol.
.macro {options.opts.asm_data_end_label} label
    .size \\label, . - \\label
.endm
"""

    nm_macros = ""
    if options.opts.asm_nonmatching_label_macro != "":
        nm_macros = f"""
# Label to signal the symbol haven't been matched yet.
.macro {options.opts.asm_nonmatching_label_macro} label, size=1
    .global \\label\\().NON_MATCHING
    .type \\label\\().NON_MATCHING, @object
    .size \\label\\().NON_MATCHING, \\size
    \\label\\().NON_MATCHING:
.endm
"""

    macros_inc = f"""\
{func_macros}
{jlabel_macro_macroinc}
{data_macros}
{nm_macros}\
"""

    if options.opts.compiler.uses_include_asm:
        # File used by original assembler
        preamble = "# This file is used by the original compiler/assembler.\n# Defines the expected assembly macros.\n"

        if options.opts.platform == "psx":
            preamble += '\n.include "gte_macros.inc"\n'

        labels_inc = f"""\
{preamble}
{func_macros}
{jlabel_macro_labelsinc}
{data_macros}
{nm_macros}\
"""
        _write(f"{directory}/labels.inc", labels_inc)

    if options.opts.platform in {"n64", "psx"}:
        gas = macros_inc
    elif options.opts.platform in {"ps2", "psp"}:
        # ps2 and psp usually use c++ mangled names, so we need to quote those
        # names when using modern gas to avoid build errors.
        # This means we can't reuse the labels.inc file.
        gas = macros_inc.replace("\\label", '"\\label"').replace(
            '"\\label"\\().NON_MATCHING', '"\\label\\().NON_MATCHING"'
        )
    elif not options.opts.is_unsupported_platform:
        log.error(f"Unknown platform '{options.opts.platform}'")
    else:
        gas = macros_inc

    if options.opts.platform == "n64":
        gas += """
# COP0 register aliases

.set Index,         $0
.set Random,        $1
.set EntryLo0,      $2
.set EntryLo1,      $3
.set Context,       $4
.set PageMask,      $5
.set Wired,         $6
.set Reserved07,    $7
.set BadVaddr,      $8
.set Count,         $9
.set EntryHi,       $10
.set Compare,       $11
.set Status,        $12
.set Cause,         $13
.set EPC,           $14
.set PRevID,        $15
.set Config,        $16
.set LLAddr,        $17
.set WatchLo,       $18
.set WatchHi,       $19
.set XContext,      $20
.set Reserved21,    $21
.set Reserved22,    $22
.set Reserved23,    $23
.set Reserved24,    $24
.set Reserved25,    $25
.set PErr,          $26
.set CacheErr,      $27
.set TagLo,         $28
.set TagHi,         $29
.set ErrorEPC,      $30
.set Reserved31,    $31

# Float register aliases

.set $fv0,          $f0
.set $fv0f,         $f1
.set $fv1,          $f2
.set $fv1f,         $f3
.set $ft0,          $f4
.set $ft0f,         $f5
.set $ft1,          $f6
.set $ft1f,         $f7
.set $ft2,          $f8
.set $ft2f,         $f9
.set $ft3,          $f10
.set $ft3f,         $f11
.set $fa0,          $f12
.set $fa0f,         $f13
.set $fa1,          $f14
.set $fa1f,         $f15
.set $ft4,          $f16
.set $ft4f,         $f17
.set $ft5,          $f18
.set $ft5f,         $f19
.set $fs0,          $f20
.set $fs0f,         $f21
.set $fs1,          $f22
.set $fs1f,         $f23
.set $fs2,          $f24
.set $fs2f,         $f25
.set $fs3,          $f26
.set $fs3f,         $f27
.set $fs4,          $f28
.set $fs4f,         $f29
.set $fs5,          $f30
.set $fs5f,         $f31
"""
    elif options.opts.platform == "psx":
        gas += '\n.include "gte_macros.inc"\n'
        write_gte_macros()

    if options.opts.generated_macro_inc_content is not None:
        gas += f"\n{options.opts.generated_macro_inc_content}\n"

    # File used by modern gas
    preamble = (
        "# This file is used by modern gas.\n# Defines the expected assembly macros\n"
    )
    gas = f"""\
# Evaluate this file only once in case it's included more than once
.ifndef _MACRO_INC_GUARD
.internal _MACRO_INC_GUARD
.set _MACRO_INC_GUARD, 1

{gas}
.endif
"""
    _write(f"{directory}/macro.inc", f"{preamble}\n{gas}")


def write_gte_macros():
    # Taken directly from https://github.com/Decompollaborate/rabbitizer/blob/-/docs/r3000gte/gte_macros.s
    # Please try to upstream any fix/update done here.
    gte_macros = """\
.ifndef .L_GTE_MACRO_INC
.L_GTE_MACRO_INC:

## GTE instruction macros
## These are meant for use with GAS and replace DMPSX

.macro cop2op fake_op, op, gbg = 0, sf = 1, mx = 0, v = 0, cv = 0, lm = 0
    cop2 \\fake_op << 20 | \\gbg << 20 | \\sf << 19 | \\mx << 17 | \\v << 15 | \\cv << 13 | \\lm << 10 | \\op
.endm

/*  RTPS    15      0x4A180001  Perspective transform */
.macro rtps
    cop2op 0x01, 0x01
.endm

/*  RTPT    23      0x4A280030  Perspective transform on 3 points */
.macro rtpt
    cop2op 0x02, 0x30
.endm

/*  DPCL    8       0x4A680029  Depth Cue Color light */
.macro dpcl
    cop2op 0x06, 0x29
.endm

/*  DPCS    8       0x4A780010  Depth Cueing */
.macro dpcs
    cop2op 0x07, 0x10
.endm

/*  DPCT    17      0x4AF8002A  Depth cue color RGB0,RGB1,RGB2 */
.macro dpct
    cop2op 0x0F, 0x2A
.endm

/*  INTPL   8       0x4A980011  Interpolation of vector and far color */
.macro intpl
    cop2op 0x09, 0x11
.endm

/*  NCS     14      0x4AC8041E  Normal color v0 */
.macro ncs
    cop2op 0x0C, 0x1E, lm = 1
.endm

/*  NCT     30      0x4AD80420  Normal color v0, v1, v2 */
.macro nct
    cop2op 0x0D, 0x20, lm = 1
.endm

/*  NCDS    19      0x4AE80413  Normal color depth cuev0 */
.macro ncds
    cop2op 0x0E, 0x13, lm = 1
.endm

/*  NCDT    44      0x4AF80416  Normal color depth cue v0, v1, v2 */
.macro ncdt
    cop2op 0x0F, 0x16, lm = 1
.endm

/*  NCCS    17      0x4B08041B  Normal color col. v0 */
.macro nccs
    cop2op 0x10, 0x1B, lm = 1
.endm

/*  NCCT    39      0x4B18043F  Normal color col.v0, v1, v2 */
.macro ncct
    cop2op 0x11, 0x3F, lm = 1
.endm

/*  CDP     13      0x4B280414  Color Depth Queue */
.macro cdp
    cop2op 0x12, 0x14, lm = 1
.endm

/*  CC      11      0x4B38041C  Color Col. */
.macro cc
    cop2op 0x13, 0x1C, lm = 1
.endm

/*  NCLIP   8       0x4B400006  Normal clipping */
.macro nclip
    cop2op 0x14, 0x06, sf = 0
.endm

/*  AVSZ3   5       0x4B58002D  Average of three Z values */
.macro avsz3
    cop2op 0x15, 0x2D
.endm

/*  AVSZ4   6       0x4B68002E  Average of four Z values */
.macro avsz4
    cop2op 0x16, 0x2E
.endm


## Instructions which take an argument
# gbg: arg is 5 bit wide
# sf : arg is 1 bit wide
# mx : arg is 2 bit wide
# v  : arg is 2 bit wide
# cv : arg is 2 bit wide
# lm : arg is 1 bit wide

/*  mvmva   8       0x4A4nnn12  Multiply vector by matrix and vector addition. */
.macro mvmva sf, mx, v, cv, lm
    cop2op 0x04, 0x12, sf = \\sf, mx = \\mx, v = \\v, cv = \\cv, lm = \\lm
.endm

/*  SQR     5       0x4AAn0428  Square of vector */
.macro sqr sf
    cop2op 0x0A, 0x28, sf = \\sf, lm = 1
.endm

/*  OP      6       0x4B7n000C  Outer Product */
.macro op sf
    cop2op 0x17, 0x0C, sf = \\sf
.endm

/*  GPF     6       0x4B9n003D  General purpose interpolation */
.macro gpf sf
    cop2op 0x19, 0x3D, sf = \\sf
.endm

/*  GPL     5       0x4BAn003E  general purpose interpolation */
.macro gpl sf
    cop2op 0x1A, 0x3E, sf = \\sf
.endm


## Convenience macros

/*  rtv0    -       0x4A486012  v0 * rotmatrix */
.macro rtv0
    # .word 0x4A486012
    mvmva       1, 0, 0, 3, 0
.endm

/*  rtv1    -       0x4A48E012  v1 * rotmatrix */
.macro rtv1
    # .word 0x4A48E012
    mvmva       1, 0, 1, 3, 0
.endm

/*  rtv2    -       0x4A496012  v2 * rotmatrix */
.macro rtv2
    # .word 0x4A496012
    mvmva       1, 0, 2, 3, 0
.endm

/*  rtir12  -       0x4A49E012  ir * rotmatrix */
.macro rtir12
    # .word 0x4A49E012
    mvmva       1, 0, 3, 3, 0
.endm

/*  rtir0   -       0x4A41E012  ir * rotmatrix */
.macro rtir0
    # .word 0x4A41E012
    mvmva       0, 0, 3, 3, 0
.endm

/*  rtv0tr  -       0x4A480012  v0 * rotmatrix + tr vector */
.macro rtv0tr
    # .word 0x4A480012
    mvmva       1, 0, 0, 0, 0
.endm

/*  rtv1tr  -       0x4A488012  v1 * rotmatrix + tr vector */
.macro rtv1tr
    # .word 0x4A488012
    mvmva       1, 0, 1, 0, 0
.endm

/*  rtv2tr  -       0x4A490012  v2 * rotmatrix + tr vector */
.macro rtv2tr
    # .word 0x4A490012
    mvmva       1, 0, 2, 0, 0
.endm

/*  rtirtr  -       0x4A498012  ir * rotmatrix + tr vector */
.macro rtirtr
    # .word 0x4A498012
    mvmva       1, 0, 3, 0, 0
.endm

/*  rtv0bk  -       0x4A482012  v0 * rotmatrix + bk vector */
.macro rtv0bk
    # .word 0x4A482012
    mvmva       1, 0, 0, 1, 0
.endm

/*  rtv1bk  -       0x4A48A012  v1 * rotmatrix + bk vector */
.macro rtv1bk
    # .word 0x4A48A012
    mvmva       1, 0, 1, 1, 0
.endm

/*  rtv2bk  -       0x4A492012  v2 * rotmatrix + bk vector */
.macro rtv2bk
    # .word 0x4A492012
    mvmva       1, 0, 2, 1, 0
.endm

/*  rtirbk  -       0x4A49A012  ir * rotmatrix + bk vector */
.macro rtirbk
    # .word 0x4A49A012
    mvmva       1, 0, 3, 1, 0
.endm

/*  ll      -       0x4A4A6412  v0 * light matrix. Lower limit result to 0 */
.macro ll
    # .word 0x4A4A6412
    mvmva       1, 1, 0, 3, 1
.endm

/*  llv0    -       0x4A4A6012  v0 * light matrix */
.macro llv0
    # .word 0x4A4A6012
    mvmva       1, 1, 0, 3, 0
.endm

/*  llv1    -       0x4A4AE012  v1 * light matrix */
.macro llv1
    # .word 0x4A4AE012
    mvmva       1, 1, 1, 3, 0
.endm

/*  llv2    -       0x4A4B6012  v2 * light matrix */
.macro llv2
    # .word 0x4A4B6012
    mvmva       1, 1, 2, 3, 0
.endm

/*  llvir   -       0x4A4BE012  ir * light matrix */
.macro llvir
    # .word 0x4A4BE012
    mvmva       1, 1, 3, 3, 0
.endm

/*  llv0tr  -       0x4A4A0012  v0 * light matrix + tr vector */
.macro llv0tr
    # .word 0x4A4A0012
    mvmva       1, 1, 0, 0, 0
.endm

/*  llv1tr  -       0x4A4A8012  v1 * light matrix + tr vector */
.macro llv1tr
    # .word 0x4A4A8012
    mvmva       1, 1, 1, 0, 0
.endm

/*  llv2tr  -       0x4A4B0012  v2 * light matrix + tr vector */
.macro llv2tr
    # .word 0x4A4B0012
    mvmva       1, 1, 2, 0, 0
.endm

/*  llirtr  -       0x4A4B8012  ir * light matrix + tr vector */
.macro llirtr
    # .word 0x4A4B8012
    mvmva       1, 1, 3, 0, 0
.endm

/*  llv0bk  -       0x4A4A2012  v0 * light matrix + bk vector */
.macro llv0bk
    # .word 0x4A4A2012
    mvmva       1, 1, 0, 1, 0
.endm

/*  llv1bk  -       0x4A4AA012  v1 * light matrix + bk vector */
.macro llv1bk
    # .word 0x4A4AA012
    mvmva       1, 1, 1, 1, 0
.endm

/*  llv2bk  -       0x4A4B2012  v2 * light matrix + bk vector */
.macro llv2bk
    # .word 0x4A4B2012
    mvmva       1, 1, 2, 1, 0
.endm

/*  llirbk  -       0x4A4BA012  ir * light matrix + bk vector */
.macro llirbk
    # .word 0x4A4BA012
    mvmva       1, 1, 3, 1, 0
.endm

/*  lc      -       0x4A4DA412  v0 * color matrix, Lower limit clamped to 0 */
.macro lc
    # .word 0x4A4DA412
    mvmva       1, 2, 3, 1, 1
.endm

/*  lcv0    -       0x4A4C6012  v0 * color matrix */
.macro lcv0
    # .word 0x4A4C6012
    mvmva       1, 2, 0, 3, 0
.endm

/*  lcv1    -       0x4A4CE012  v1 * color matrix */
.macro lcv1
    # .word 0x4A4CE012
    mvmva       1, 2, 1, 3, 0
.endm

/*  lcv2    -       0x4A4D6012  v2 * color matrix */
.macro lcv2
    # .word 0x4A4D6012
    mvmva       1, 2, 2, 3, 0
.endm

/*  lcvir   -       0x4A4DE012  ir * color matrix */
.macro lcvir
    # .word 0x4A4DE012
    mvmva       1, 2, 3, 3, 0
.endm

/*  lcv0tr  -       0x4A4C0012  v0 * color matrix + tr vector */
.macro lcv0tr
    # .word 0x4A4C0012
    mvmva       1, 2, 0, 0, 0
.endm

/*  lcv1tr  -       0x4A4C8012  v1 * color matrix + tr vector */
.macro lcv1tr
    # .word 0x4A4C8012
    mvmva       1, 2, 1, 0, 0
.endm

/*  lcv2tr  -       0x4A4D0012  v2 * color matrix + tr vector */
.macro lcv2tr
    # .word 0x4A4D0012
    mvmva       1, 2, 2, 0, 0
.endm

/*  lcirtr  -       0x4A4D8012  ir * color matrix + tr vector */
.macro lcirtr
    # .word 0x4A4D8012
    mvmva       1, 2, 3, 0, 0
.endm

/*  lev0bk  -       0x4A4C2012  v0 * color matrix + bk vector */
.macro lev0bk
    # .word 0x4A4C2012
    mvmva       1, 2, 0, 1, 0
.endm

/*  lev1bk  -       0x4A4CA012  v1 * color matrix + bk vector */
.macro lev1bk
    # .word 0x4A4CA012
    mvmva       1, 2, 1, 1, 0
.endm

/*  lev2bk  -       0x4A4D2012  v2 * color matrix + bk vector */
.macro lev2bk
    # .word 0x4A4D2012
    mvmva       1, 2, 2, 1, 0
.endm

/*  leirbk  -       0x4A4DA012  ir * color matrix + bk vector */
.macro leirbk
    # .word 0x4A4DA012
    mvmva       1, 2, 3, 1, 0
.endm

/*  sqr12   -       0x4AA80428  square of ir    1,19,12 */
# .macro sqr12
#     # .word 0x4AA80428
#     sqr         1
# .endm

/*  sqr0    -       0x4AA00428  square of ir    1,31, 0 */
# .macro sqr0
#     # .word 0x4AA00428
#     sqr         0
# .endm

/*  op12    -       0x4B78000C  outer product   1,19,12 */
.macro op12
    # .word 0x4B78000C
    op          1
.endm

/*  op0     -       0x4B70000C  outer product   1,31, 0 */
.macro op0
    # .word 0x4B70000C
    op          0
.endm

/*  gpf12   -       0x4B98003D  general purpose interpolation   1,19,12 */
.macro gpf12
    # .word 0x4B98003D
    gpf         1
.endm

/*  gpf0    -       0x4B90003D  general purpose interpolation   1,31, 0 */
.macro gpf0
    # .word 0x4B90003D
    gpf         0
.endm

/*  gpl12   -       0x4BA8003E  general purpose interpolation   1,19,12 */
.macro gpl12
    # .word 0x4BA8003E
    gpl         1
.endm

/*  gpl0    -       0x4BA0003E  general purpose interpolation   1,31, 0 */
.macro gpl0
    # .word 0x4BA0003E
    gpl         0
.endm

.endif
"""

    directory = options.opts.generated_asm_macros_directory.as_posix()
    _write(f"{directory}/gte_macros.inc", gte_macros)
